"""Tools for VASP input and output"""
from casm.vasp.io.io import \
    VASP_INPUT_FILE_LIST,\
    DEFAULT_VASP_MOVE_LIST,\
    DEFAULT_VASP_COPY_LIST,\
    DEFAULT_VASP_REMOVE_LIST,\
    VaspIOError,\
    job_complete,\
    get_incar_tag,\
    set_incar_tag,\
    ionic_steps,\
    write_potcar,\
    write_stopcar,\
    write_vasp_input
from casm.vasp.io.incar import \
    VASP_TAG_INT_LIST,\
    VASP_TAG_FLOAT_LIST,\
    VASP_TAG_BOOL_LIST,\
    VASP_TAG_SITEF_LIST,\
    VASP_TAG_SPECF_LIST,\
    VASP_TAG_SPECI_LIST,\
    VASP_TAG_STRING_LIST,\
    VASP_TAG_LIST,\
    IncarError,\
    Incar
from casm.vasp.io.kpoints import KpointsError, Kpoints
from casm.vasp.io.outcar import OutcarError, Outcar
from casm.vasp.io.oszicar import OszicarError, Oszicar
from casm.vasp.io.poscar import Site, PoscarError, Poscar
from casm.vasp.io.species import SpeciesError, SpeciesDict, IndividualSpecies,\
    species_settings, write_species_settings
from casm.vasp.io.vaspio import VaspIO
from casm.vasp.io.vasprun import VasprunError, Vasprun
from casm.vasp.io.orbital_occupation import OrbitalOccupationError, OrbitalOccupation, write_occupations
from casm.vasp.io.attribute_classes import CmagspinAttr
__all__ = [
    'VASP_INPUT_FILE_LIST', 'DEFAULT_VASP_MOVE_LIST', 'DEFAULT_VASP_COPY_LIST',
    'DEFAULT_VASP_REMOVE_LIST', 'VaspIOError', 'job_complete', 'get_incar_tag',
    'set_incar_tag', 'ionic_steps', 'write_potcar', 'write_stopcar',
    'write_vasp_input', 'VASP_TAG_INT_LIST', 'VASP_TAG_FLOAT_LIST',
    'VASP_TAG_BOOL_LIST', 'VASP_TAG_SITEF_LIST', 'VASP_TAG_SPECF_LIST',
    'VASP_TAG_SPECI_LIST', 'VASP_TAG_STRING_LIST', 'VASP_TAG_LIST',
    'IncarError', 'Incar', 'KpointsError', 'Kpoints', 'OutcarError', 'Outcar',
    'OszicarError', 'Oszicar', 'Site', 'PoscarError', 'Poscar', 'SpeciesError',
    'SpeciesDict', 'IndividualSpecies', 'species_settings',
    'write_species_settings', 'VaspIO', 'VasprunError', 'Vasprun',
    'OrbitalOccupationError', 'OrbitalOccupation', 'write_occupations',
    'CmagspinAttr'
]



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import re
import six

# Note: This could use double-checking

# Schoenflies to Hermann-Mauguin
symmap = {
    'C1': '1',
    'Ci': '-1',
    'S2': '-1',
    'C2': '2',
    'Cs': 'm',
    'C1h': 'm',
    'C2h': '2/m',
    'D2': '222',
    'V': '222',
    'C2v': 'mm2',
    'D2h': 'mmm',
    'Vh': 'mmm',
    'C4': '4',
    'S4': '-4',
    'C4h': '4/m',
    'D4': '422',
    'C4v': '4mm',
    'D2d': '-42m',
    'Vd': '-42m',
    'D4h': '4/mmm',
    'C3': '3',
    'S6': '-3',
    'C3i': '-3',
    'D3': '32',
    'C3v': '3m',
    'D3d': '-3m',
    'C6': '6',
    'C3h': '-6',
    'C6h': '6/m',
    'D6': '622',
    'C6v': '6mm',
    'D3h': '-6m2',
    'D6h': '6/mmm',
    'T': '23',
    'Th': 'm-3',
    'O': '432',
    'Td': '-43m',
    'Oh': 'm-3m',
}

# lattice system
latsysmap = {
    'Ci': 'triclinic',
    'S2': 'triclinic',
    'C2h': 'monoclinic',
    'D2h': 'orthorhombic',
    'D4h': 'tetragonal',
    'D3d': 'rhombohedral',
    'D6h': 'hexagonal',
    'Oh': 'cubic',
}

# crystal system to Schoenflies
xtalsysmap = {
    'triclinic': ['C1', 'Ci', 'S2'],
    'monoclinic': ['C2', 'Cs', 'C1h', 'C2h'],
    'orthorhombic': ['D2', 'V', 'C2v', 'D2h', 'Vh'],
    'tetragonal': ['C4', 'S4', 'C4h', 'D4', 'C4v', 'D2d', 'Vd', 'D4h'],
    'trigonal': ['C3', 'S6', 'C3i', 'D3', 'C3v', 'D3d'],
    'hexagonal': ['C6', 'C3h', 'C6h', 'D6', 'C6v', 'D3h', 'D6h'],
    'cubic': ['T', 'Th', 'O', 'Td', 'Oh']
}

# crystal family to Schoenflies
xtalfamilymap = {
    'triclinic': ['C1', 'Ci', 'S2'],
    'monoclinic': ['C2', 'Cs', 'C1h', 'C2h'],
    'orthorhombic': ['D2', 'V', 'C2v', 'D2h', 'Vh'],
    'tetragonal': ['C4', 'S4', 'C4h', 'D4', 'C4v', 'D2d', 'Vd', 'D4h'],
    'hexagonal': [
        'C3', 'S6', 'C3i', 'D3', 'C3v', 'D3d', 'C6', 'C3h', 'C6h', 'D6', 'C6v',
        'D3h', 'D6h'
    ],
    'cubic': ['T', 'Th', 'O', 'Td', 'Oh']
}

# Schoenflies to International Tables space group number (range)
space_group_number_map = {
    'C1': '1',
    'Ci': '2',
    'S2': '2',
    'C2': '3:5',
    'Cs': '6:9',
    'C1h': '6:9',
    'C2h': '10:15',
    'D2': '16:24',
    'V': '16:24',
    'C2v': '25:46',
    'D2h': '47:74',
    'Vh': '47:74',
    'C4': '75:80',
    'S4': '81:82',
    'C4h': '83:88',
    'D4': '89:98',
    'C4v': '99:110',
    'D2d': '111:122',
    'Vd': '111:122',
    'D4h': '123:142',
    'C3': '143:146',
    'S6': '147:148',
    'C3i': '149:155',
    'D3': '149:155',
    'C3v': '156:161',
    'D3d': '162:167',
    'C6': '168:173',
    'C3h': '174',
    'C6h': '175:176',
    'D6': '177:182',
    'C6v': '183:186',
    'D3h': '187:190',
    'D6h': '191:194',
    'T': '195:199',
    'Th': '200:206',
    'O': '207:214',
    'Td': '215:220',
    'Oh': '221:230',
}


def lattice_symmetry(stdout):
    """
    Parse output from 'casm sym' for lattice point group, return Schoenflies form
    """
    for line in stdout.split('\n'):
        if re.search('Lattice point group is:', line):
            return line.split()[-1]
    return None


def hm_symmetry(schoenflies_symbol):
    """
    Determine Hermann-Mauguin symbol from Schoenflies symbol
    """
    return symmap[schoenflies_symbol]


def lattice_system(schoenflies_symbol):
    """
    Determine lattice system from lattice symmetry
    """
    return latsysmap[schoenflies_symbol]


def crystal_symmetry(stdout):
    """
    Parse output from 'casm sym' for crystal point group, return Schoenflies form
    """
    for line in stdout.split('\n'):
        if re.search('Crystal point group is:', line):
            return line.split()[-1]
    return None


def crystal_system(schoenflies_symbol):
    """ 
    Determine crystal system from crystal symmetry
    """
    for key, val in six.iteritems(xtalsysmap):
        if schoenflies_symbol in val:
            return key
    return None


def crystal_family(schoenflies_symbol):
    """ 
    Determine crystal family from crystal symmetry
    """
    for key, val in six.iteritems(xtalfamilymap):
        if schoenflies_symbol in val:
            return key
    return None



"""A Python package for interacting with VASP"""
from casm.vasp.error import \
    VaspError,\
    VaspWarning,\
    continue_job,\
    IbzkptError,\
    FEXCFError,\
    SubSpaceMatrixError,\
    InisymError,\
    SgrconError,\
    WavecarError,\
    NbandsError,\
    NoConvergeError,\
    FreezeError,\
    error_check,\
    crash_check
from casm.vasp.run import \
    complete_job,\
    run
from casm.vasp.relax import Relax
from casm.vasp.converge import Converge
__all__ = [
    'io', 'VaspError', 'VaspWarning', 'continue_job', 'IbzkptError',
    'FEXCFError', 'SubSpaceMatrixError', 'InisymError', 'SgrconError',
    'WavecarError', 'NbandsError', 'NoConvergeError', 'FreezeError',
    'error_check', 'crash_check', 'complete_job', 'run', 'Relax', 'Converge'
]



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import json
import os
import math
import tempfile
import warnings
from os.path import join
from string import ascii_lowercase

import numpy as np
import six

from casm.project import syminfo
from casm.api import API, casm_command, casm_capture
from casm.api.api import _project_path as _api_project_path


def project_path(dir=None):
    """
    Crawl up from dir to find '.casm'. If found returns the directory containing the '.casm' directory.
    If not found, return None.

    Args:
    If dir == None, set to os.getcwd()
    """
    return _api_project_path(dir=dir)


class ClexDescription(object):
    """
    Settings for a cluster expansion

    Attributes
    ----------

      name: str
        Cluster expansion name

      property: str
        Name of the property being cluster expanded

      calctype: str
        Calctype name

      ref: str
        Reference state name

      bset: str
        Basis set

      eci: str
        ECI set name

    """
    def __init__(self, name, property, calctype, ref, bset, eci):
        self.name = name
        self.property = property
        self.calctype = calctype
        self.ref = ref
        self.bset = bset
        self.eci = eci


class ProjectSettings(object):
    """
    Settings for a CASM project

    Attributes:
      path: Path to CASM project
      data: Dict storing contents of project_settings.json

    """
    def __init__(self, path=None):
        """
        Construct a CASM ProjectSettings representation.

        Args:
            path: path to CASM project (Default=None, uses project containing current directory).

        """
        if project_path(path) is None:
            if path is None:
                raise Exception("No CASM project found using " + os.getcwd())
            else:
                raise Exception("No CASM project found using " + path)
        self.path = project_path(path)
        dir = DirectoryStructure(self.path)
        self.data = json.load(open(dir.project_settings()))

        d = self.data["cluster_expansions"][self.data["default_clex"]]
        self._default_clex = ClexDescription(d["name"], d["property"],
                                             d["calctype"], d["ref"],
                                             d["bset"], d["eci"])

        self._clex = [
            ClexDescription(d[1]["name"], d[1]["property"], d[1]["calctype"],
                            d[1]["ref"], d[1]["bset"], d[1]["eci"])
            for d in six.iteritems(self.data["cluster_expansions"])
        ]

        d = self.data["cluster_expansions"].get("formation_energy", None)
        self._formation_energy_clex = None
        if d is not None:
            self._formation_energy_clex = ClexDescription(
                d["name"], d["property"], d["calctype"], d["ref"], d["bset"],
                d["eci"])

    # -- Accessors --

    @property
    def cluster_expansions(self):
        return self._clex

    @property
    def default_clex(self):
        return self._default_clex

    @property
    def formation_energy_clex(self):
        return self._formation_energy_clex


class DirectoryStructure(object):
    """Standard file and directory locations for a CASM project"""
    def __init__(self, path=None):
        """
        Construct a CASM Project DirectoryStructure representation.

        Args:
            path: path to CASM project (Default=None, uses project containing current directory).

        """
        if project_path(path) is None:
            if path is None:
                raise Exception("No CASM project found using " + os.getcwd())
            else:
                raise Exception("No CASM project found using " + path)
        self.path = project_path(path)
        self.__casm_dir = ".casm"
        self.__casmdb_dir = "jsonDB"
        self.__bset_dir = "basis_sets"
        self.__calc_dir = "training_data"
        self.__set_dir = "settings"
        self.__sym_dir = "symmetry"
        self.__clex_dir = "cluster_expansions"

    # ** Query filesystem **

    def all_bset(self):
        """Check filesystem directory structure and return list of all basis set names"""
        return self.__all_settings("bset", join(self.path, self.__bset_dir))

    def all_calctype(self):
        """Check filesystem directory structure and return list of all calctype names"""
        return self.__all_settings(
            "calctype", join(self.path, self.__calc_dir, self.__set_dir))

    def all_ref(self, calctype):
        """Check filesystem directory structure and return list of all ref names for a given calctype"""
        return self.__all_settings("ref", self.calc_settings_dir(calctype))

    def all_clex_name(self):
        """Check filesystem directory structure and return list of all cluster expansion names"""
        return self.__all_settings("clex", join(self.path, self.__clex_dir))

    def all_eci(self, property, calctype, ref, bset):
        """Check filesystem directory structure and return list of all eci names"""
        return self.__all_settings(
            "eci",
            join(self.path, self.__clex_dir, self.__clex_name(property),
                 self.__calctype(calctype), self.__ref(ref),
                 self.__bset(bset)))

    # ** File and Directory paths **

    # -- Project directory --------

    def root_dir(self):
        """Return casm project directory path"""
        return self.path

    def prim(self):
        """Return prim.json path"""
        return join(self.path, "prim.json")

    # -- Hidden .casm directory --------

    def casm_dir(self):
        """Return hidden .casm dir path"""
        return join(self.path, self.__casm_dir)

    def casmdb_dir(self):
        """Return .casm/jsonDB path"""
        return join(self.casm_dir, self.__casmdb_dir)

    def project_settings(self):
        """Return project_settings.json path"""
        return join(self.casm_dir(), "project_settings.json")

    def scel_list(self, scelname):
        """Return master scel_list.json path"""
        return join(self.casmdb_dir(), "scel_list.json")

    def config_list(self):
        """Return master config_list.json file path"""
        return join(self.casm_dbdir(), "config_list.json")

    def master_selection(self, type):
        """Return location of MASTER selection file

        Arguments
        ---------
        type: str
            One of "config" or "scel"
        """
        querydir = join(self.casmdb_dir(), "query")
        if type == "config":
            return join(querydir, "Configuration", "master_selection")
        elif type == "scel":
            return join(querydir, "Supercell", "master_selection")
        else:
            raise Exception("Unsupported type: " + str(type))

    # -- Symmetry --------

    def symmetry_dir(self):
        """Return symmetry directory path"""
        return join(self.path, self.sym_dir)

    def lattice_point_group(self):
        """Return lattice_point_group.json path"""
        return join(self.symmetry_dir(), "lattice_point_group.json")

    def factor_group(self):
        """Return factor_group.json path"""
        return join(self.symmetry_dir(), "factor_group.json")

    def crystal_point_group(self):
        """Return crystal_point_group.json path"""
        return join(self.symmetry_dir(), "crystal_point_group.json")

    # -- Basis sets --------

    def bset_dir(self, clex):
        """Return path to directory contain basis set info"""
        return join(self.path, self.__bset_dir, self.__bset(clex.bset))

    def bspecs(self, clex):
        """Return basis function specs (bspecs.json) file path"""
        return join(self.bset_dir(clex), "bspecs.json")

    def clust(self, clex):
        """Returns path to the clust.json file"""
        return join(self.bset_dir(clex), "clust.json")

    def basis(self, clex):
        """Returns path to the basis.json file"""
        return join(self.bset_dir(clex), "basis.json")

    def clexulator_dir(self, clex):
        """Returns path to directory containing global clexulator"""
        return join(self.bset_dir(clex))

    def clexulator_src(self, project, clex):
        """Returns path to global clexulator source file"""
        return join(self.bset_dir(clex), (project + "_Clexulator.cc"))

    def clexulator_o(self, project, clex):
        """Returns path to global clexulator.o file"""
        return join(self.bset_dir(clex), (project + "_Clexulator.o"))

    def clexulator_so(self, project, clex):
        """Returns path to global clexulator so file"""
        return join(self.bset_dir(clex), (project + "_Clexulator.so"))

    # -- Calculations and reference --------

    def settings_path_crawl(self, filename, configname, clex, calc_subdir=""):
        """
        Returns the path to the first file named 'filename' found in the settings
        directories.

        Searches:
          1) self.configuration_calc_settings_dir(configname, clex)
          2) self.supercell_calc_settings_dir(scelname, clex)
          3) self.calc_settings_dir(clex)
          DirectoryStructure.configuration_calc_settings_dir(configname, clex)Crawl casm directory structure starting at configdir and moving upwards

        Returns None if file named 'filename' not found in any of the three directories.


        Arguments
        ---------
          filename: str
            The name of the file being searched for

          configname: str
            The name of the configuration

          clex: a casm.project.ClexDescription instance
            Used to specify the calctype to find settings for


        Returns
        ---------
          filepath: str or None
            The path to the first file named 'filename' found in the settings
            directories, or None if not found.

        """
        filepath = join(
            self.configuration_calc_settings_dir(configname, clex,
                                                 calc_subdir), filename)
        if os.path.exists(filepath):
            return filepath

        scelname = configname.split('/')[0]
        filepath = join(
            self.supercell_calc_settings_dir(scelname, clex, calc_subdir),
            filename)
        if os.path.exists(filepath):
            return filepath

        filepath = join(self.calc_settings_dir(clex), filename)
        if os.path.exists(filepath):
            return filepath

        return None

    def supercell_dir(self, scelname, calc_subdir=""):
        """Return supercell directory path (scelname has format SCELV_A_B_C_D_E_F)"""
        return join(self.path, self.__calc_dir, calc_subdir, scelname)

    def configuration_dir(self, configname, calc_subdir=""):
        """Return configuration directory path (configname has format SCELV_A_B_C_D_E_F/I)"""
        return join(self.path, self.__calc_dir, calc_subdir, configname)

    def POS(self, configname, calc_subdir=""):
        """Return path to POS file"""
        return join(self.configuration_dir(configname, calc_subdir), "POS")

    def config_json(self, configname, calc_subdir=""):
        """Return path to structure.json file"""
        return join(self.configuration_dir(configname, calc_subdir),
                    "structure.json")

    def structure_json(self, configname, calc_subdir=""):
        """Return path to structure.json file"""
        return join(self.configuration_dir(configname, calc_subdir),
                    "structure.json")

    def calctype_dir(self, configname, clex, calc_subdir=""):
        """Return calctype directory path (e.g. training_data/$(calc_subdir)/SCEL_...../0/calctype.default"""
        return join(self.configuration_dir(configname, calc_subdir),
                    self.__calctype(clex.calctype))

    def calc_settings_dir(self, clex):
        """Return calculation settings directory path, for global settings from clex"""
        return join(self.path, self.__calc_dir, self.__set_dir,
                    self.__calctype(clex.calctype))

    def calctype_settings_dir(self, calctype):
        """Return calculation settings directory path, for global settings from calctype"""
        return join(self.path, self.__calc_dir, self.__set_dir,
                    self.__calctype(calctype))

    def supercell_calc_settings_dir(self, scelname, clex, calc_subdir=""):
        """Return calculation settings directory path, for supercell specific settings"""
        return join(self.supercell_dir(scelname, calc_subdir), self.__set_dir,
                    self.__calctype(clex.calctype))

    def configuration_calc_settings_dir(self,
                                        configname,
                                        clex,
                                        calc_subdir=""):
        """Return calculation settings directory path, for configuration specific settings"""
        return join(self.configuration_dir(configname, calc_subdir),
                    self.__set_dir, self.__calctype(clex.calctype))

    def calculated_properties(self, configname, clex, calc_subdir=""):
        """Return calculated properties file path"""
        return join(self.configuration_dir(configname, calc_subdir),
                    self.__calctype(clex.calctype), "properties.calc.json")

    def ref_dir(self, clex):
        """Return calculation reference settings directory path, for global settings"""
        return join(self.calc_settings_dir(clex.calctype),
                    self.__ref(clex.ref))

    def composition_axes(self):
        """Return composition axes file path"""
        return join(self.casm_dir(), "composition_axes.json")

    def chemical_reference(self, clex):
        """Return chemical reference file path"""
        return join(self.ref_dir(clex), "chemical_reference.json")

    # -- Cluster expansions --------

    def property_dir(self, clex):
        """Returns path to eci directory"""
        return join(self.path, self.__clex_dir,
                    self.__clex_name(clex.property))

    def eci_dir(self, clex):
        """
      Returns path to eci directory

      Arguments
      ---------
        clex: a casm.project.ClexDescription instance
          Specifies the cluster expansion to get the eci directory for

      Returns
      -------
        p: str
          Path to the eci directory
      """
        return join(self.property_dir(clex), self.__calctype(clex.calctype),
                    self.__ref(clex.ref), self.__bset(clex.bset),
                    self.__eci(clex.eci))

    def eci(self, clex):
        """
      Returns path to eci.json

      Arguments
      ---------
        clex: a casm.project.ClexDescription instance
          Specifies the cluster expansion to get the eci.json for

      Returns
      -------
        p: str
          Path to the eci directory
      """
        return join(self.eci_dir(clex), "eci.json")

    # private:

    def __bset(self, bset):
        return "bset." + bset

    def __calctype(self, calctype):
        return "calctype." + calctype

    def __ref(self, ref):
        return "ref." + ref

    def __clex_name(self, clex_name):
        return "clex." + clex_name

    def __eci(self, eci):
        return "eci." + eci

    def __all_settings(self, pattern, location):
        """
      Find all directories at 'location' that match 'pattern.something'
      and return a std::vector of the 'something'
      """

        all = []
        pattern += "."

        # get all
        if not os.path.exists(location):
            return all

        for item in os.listdir(location):
            if os.path.isdir(os.path.join(
                    location, item)) and item[:len(pattern)] == pattern:
                all.append(item[len(pattern):])
        return sorted(all)


class Project(object):
    """The Project class contains information about a CASM project

    Attributes
    ----------

      path: str
        Path to project root directory

      name: str
        Project name

      settings: casm.project.ProjectSettings instance
        Contains project settings

      dir: casm.project.DirectoryStructure instance
        Provides file and directory locations within the project

      prim: casm.project.Prim instance
        Represents the primitive crystal structure

      composition_axes: casm.project.CompositionAxes or None
        Currently selected composition axes, or None

      all_composition_axes: dict(str:casm.project.CompositionAxes)
        Dict containing name:CompositionAxes pairs, including both standard and custom composition axes

      verbose: bool
        How much to print to stdout

      out: str or None
        Contains last output of Project.command calls if capturing output. Use Project.command_options to set options.

      err: str or None
        Contains last error output of Project.command calls if capturing output. Use Project.command_options to set options.

      code: int or None
        Contains last return code of Project.command calls.

    """
    def __init__(self, path=None, verbose=True):
        """
        Construct a CASM Project representation.

        Arguments
        ----------

          path: str, optional, default=None
            Path to project root directory. Default=None uses project containing
            current working directory

          verbose: bool, optional, default=True
            How much to print to stdout

        """

        # will hold a ctypes.c_void_p when loading CASM project into memory
        self._ptr = None

        # will keep a casm.API instance
        self._api = None

        # set path to this CASM project
        if project_path(path) is None:
            if path is None:
                raise Exception("No CASM project found using " + os.getcwd())
            else:
                raise Exception("No CASM project found using " + path)

        self.path = project_path(path)
        self.__refresh()

        self.verbose = verbose
        self._streamptr = None
        self._errstreamptr = None

        # set default command output options
        self.out = None
        self.err = None
        self.code = None
        self.command_options()

    def __del__(self):
        self.__unload()

    def __load(self):
        """
        Explicitly load CASM project into memory.
        """
        if self._ptr is None:
            self._api = API()
            if self.verbose:
                self._streamptr = self._api.stdout()
            else:
                self._streamptr = self._api.nullstream()

            if self.verbose:
                self._errstreamptr = self._api.stderr()
            else:
                self._errstreamptr = self._api.nullstream()

            self._ptr = self._api.primclex_new(self.path, self._streamptr,
                                               self._errstreamptr)

    def __unload(self):
        """
        Explicitly unload CASM project from memory.
        """
        if self._ptr is not None:
            self._api.primclex_delete(self._ptr)
            self._ptr = None

    def __refresh(self):
        """
        Reload self.settings and self.dir

        Use this after adding or modifying files in the CASM project but no
        special call to refresh PrimClex properties is required
        """
        self.dir = DirectoryStructure(self.path)
        self.settings = ProjectSettings(self.path)
        self._prim = None
        self.all_composition_axes = {}
        if os.path.exists(self.dir.composition_axes()):
            with open(self.dir.composition_axes(), 'r') as f:
                data = json.load(f)
                if "possible_axes" in data:
                    for key, val in six.iteritems(data["possible_axes"]):
                        self.all_composition_axes[key] = CompositionAxes(
                            key, val)
                if "custom_axes" in data:
                    for key, val in six.iteritems(data["custom_axes"]):
                        self.all_composition_axes[key] = CompositionAxes(
                            key, val)
                self.composition_axes = None
                if "current_axes" in data:
                    self.composition_axes = self.all_composition_axes[
                        data["current_axes"]]

    @property
    def prim(self):
        if self._prim is None:
            self._prim = Prim(self)
        return self._prim

    @property
    def name(self):
        return self.settings.data['name']

    def refresh(self,
                read_settings=False,
                read_composition=False,
                read_chem_ref=False,
                read_configs=False,
                clear_clex=False):
        """
        Refresh PrimClex properties to reflect changes to CASM project files.
        """
        if read_settings:
            self.__refresh()
        if self._ptr is not None:
            self._api.primclex_refresh(self.data(), self._streamptr,
                                       self._errstreamptr, read_settings,
                                       read_composition, read_chem_ref,
                                       read_configs, clear_clex)

    def data(self):
        """
        Returns a 'ctypes.c_void_p' that points to a CASM project. (PrimClex)
        """
        self.__load()
        return self._ptr

    def command_options(self,
                        capture=True,
                        print_output=True,
                        combine_output=True):
        """
        Set options for Project.command

        Arguments
        ---------
        capture: bool
            If True, capture stdout, stderr, and return code in self.out and self.err, and self.code. If False, will print to stdout and stderr directrly (if self.verbose==True), or to nullstream (if self.verbose==False).

        print_ouput: bool
            If print_output==True and capture==True, then print self.out after executing.

        combine_output: bool
            If combine_output==true, err stream is set equal to out stream.

        """
        self._capture = capture
        self._print_output = print_output
        self._combine_output = combine_output

    def command(self, args):
        """
        Execute a command via the c api, writing output to stdout/stderr.

        Args:
          args: A string containing the command to be executed.
            Ex: "select --set-on -o /abspath/to/my_selection"

        Returns:
          returncode: The returncode of the command via the CASM C API.

        """
        # this also ensures self._api is not None
        data = self.data()

        if self._capture:
            if self._combine_output:
                self.out, self.code = self.capture(args, combine_output=True)
            else:
                self.out, self.err, self.code = self.capture(args)
            if self._print_output:
                print(self.out)
        else:
            self.code = self._api.capi(args, self.data(), self.path,
                                       self._streamptr, self._errstreamptr)
            self.__refresh()

    def capture(self, args, combine_output=False):
        """
        Execute a command via the c api and store stdout/stderr result as str.

        Args:
          args: A string containing the command to be executed.
            Ex: "select --set-on -o /abspath/to/my_selection"

        Returns
        -------
          (stdout, stderr, returncode): The result of running the command via the
            command line iterface. 'stdout' and 'stderr' are in text type ('unicode'/'str'). If
            'combine_output' is True, then returns (combined_output, returncode).

        """
        # this also ensures self._api is not None
        data = self.data()

        # construct stringstream objects to capture stdout, debug, stderr
        ss = self._api.ostringstream_new()
        if combine_output:
            ss_err = ss
        else:
            ss_err = self._api.ostringstream_new()

        returncode = self._api.capi(args, self.data(), self.path, ss, ss_err)

        # copy strings and delete stringstreams
        stdout = self._api.ostringstream_to_str(ss)
        self._api.ostringstream_delete(ss)

        if combine_output:
            res = (stdout.decode('utf-8'), returncode)
        else:
            stderr = self._api.ostringstream_to_str(ss_err)
            self._api.ostringstream_delete(ss_err)

            res = (stdout.decode('utf-8'), stderr.decode('utf-8'), returncode)

        self.__refresh()
        return res

    @classmethod
    def init(cls,
             root=None,
             prim_path=None,
             prim_str=None,
             verbose=True,
             subproject=False,
             config_selection_path=None,
             confignames=None,
             dofs=None,
             relaxed=False,
             include_va=False,
             as_molecules=False):
        """ Calls `casm init` to create a new CASM project in the given directory

        Arguments
        ---------

          root: str (optional, default=os.getcwd())
            A string giving the path to the root directory of the new CASM
            project. Raises if a project already exists at this location.

          prim_path: str (optional, default="prim.json")
            A string giving the path to a `prim.json` file to initialize the
            CASM project with.

          prim_str: str (optional, default=None)
            A string with the `prim.json` file contents used to initialize the
            CASM project with.

          verbose: bool (optional, default=True)
            Passed to casm.project.Project constructor. How much to print to
            stdout.

          subproject: bool (optional, default=False)
            Initialize a project in sub-directory of an existing project. After
            initialization, commands executed below the sub-directory will act
            on the sub-directory project; commmands executed above the sub-
            directory act on the original project.

          config_selection_path: str (optional, default=None)
            If not None, initialize subprojects using existing configurations
            from the given selection file. Options are: {'<filename>', 'MASTER'
            (default), 'ALL', 'NONE', 'EMPTY', 'CALCULATED'}.

          confignames: list of str (optional, default=None)
            If not None, initialize subprojects using the configurations
            specified by name.

          dofs: list of str (optional, deafult=None)
            One or more DoF types to use casm init with, such as 'disp' or
            'EAstrain'.

          relaxed: bool (optional, default=False)
            Utilize relaxed coordinates for writing configuration-specific
            prim.json files.

          include_va: bool (optional, default=False)
            Print sites that can only be vacancies; otherwise these sites are
            excluded.

          as_molecules: bool (optional, default=False)
            Keep multi-atom species as molecules. By default, multi-atom
            species are split into constituent atoms.

        Returns
        -------
          proj: A casm.project.Project instance for the new CASM project. The new project has composition axes calculated and the first axes choice selected.

        Raises
        ------
          An exception is raised if a new project could not be initialized. This could be due to an already existing project, bad or missing input file, or other cause.

        """
        if root is None:
            root = os.getcwd()
        root = str(root)

        if os.path.exists(root) and subproject is False:
            if project_path(root) is not None:
                raise Exception("A CASM project already exists at " + root)

        os.makedirs(root, exist_ok=True)

        tmpdir = None

        if prim_path is not None:
            prim_path = str(prim_path)
        elif prim_path is None and prim_str is not None:
            # write "prim.json" file
            tmpdir = tempfile.TemporaryDirectory()
            prim_path = os.path.join(tmpdir.name, "prim.json")
            with open(prim_path, 'w') as f:
                f.write(prim_str)

        def raise_on_fail(output, returncode):
            if returncode != 0:
                print(output)
                raise Exception("Could not initialize the project")

        args = "init --path=" + str(root)
        if prim_path is not None:
            args += " --prim=" + str(prim_path)
        if subproject:
            args += " --sub"
        if config_selection_path is not None or confignames is not None:
            args += " --write-prim"
            if config_selection_path is not None:
                args += " -c " + str(config_selection_path)
            if confignames is not None:
                args += " --confignames " + " ".join(confignames)
        if dofs is not None:
            args += " --dofs " + " ".join(dofs)
        if relaxed is True:
            args += " --relaxed"
        if include_va is True:
            args += " --include_va"
        if as_molecules is True:
            args += " --as-molecules"
        raise_on_fail(*casm_capture(args, combine_output=True))
        proj = Project(root, verbose=verbose)
        raise_on_fail(*proj.capture("composition --calc", combine_output=True))
        raise_on_fail(
            *proj.capture("composition --select 0", combine_output=True))
        return proj


class Prim(object):
    """The Primitive Crystal Structure

    Attributes
    ----------

        proj: casm.Project
          the CASM project the Prim belongs to

        lattice_matrix: np.array of shape (3, 3)
          lattice vectors as matrix columns

        lattice_parameters: dict
          Lattice parameters and angles (in degrees), as:
            {'a':a, 'b':b, 'c':c, 'alpha':alpha, 'beta':beta, 'gamma':gamma}

        basis: List(dict)
          crystal basis, as read directly from prim.json (format may change)

        coordinate_mode: str
          crystal basis coordinate_mode, as read directly from prim.json (format
          may change)

        lattice_symmetry_s: str
          lattice point group, in Schoenflies notation

        lattice_symmetry_hm: str
          lattice point group, in Hermann-Mauguin notation

        lattice_system: str
          lattice system name, ('cubic', 'hexagonal', 'rhombohedral', etc.)

        crystal_symmetry_s: str
          crystal point group, in Schoenflies notation

        crystal_symmetry_hm: str
          crystal point group, in Hermann-Mauguin notation

        crystal_system: str
          crystal system name, ('cubic', 'hexagonal', 'trigonal', etc.)

        crystal_family: str
          crystal family name, ('cubic', 'hexagonal', etc.)

        space_group_number: str
          range of possible space group number
    """

    # TODO: update prim composition info
    #
    # components: List[str]
    #   occupational components
    #
    # elements: List[str]
    #   all allowed elements
    #
    # n_independent_compositions: int
    #   number of independent composition axes
    #
    # degrees_of_freedom: List[str]
    #   allowed degrees of freedom, from:
    #     'occupation'
    def __init__(self, proj):
        """
        Construct a CASM Prim

        Arguments
        ---------

          proj: casm.Project, optional, default=Project containing the current working directory
            the CASM project the Prim belongs to

        """
        if proj == None:
            proj = Project()
        elif not isinstance(proj, Project):
            raise Exception(
                "Error constructing Prim: proj argument is not a CASM Project")
        self.proj = proj

        # raw prim.json (for some properties not yet supported in the API)
        with open(self.proj.dir.prim()) as f:
            raw_prim = json.load(f)
        self.lattice_matrix = np.array(raw_prim['lattice_vectors']).transpose()
        self.basis = raw_prim['basis']
        self.coordinate_mode = raw_prim['coordinate_mode']

        def _angle(a, b):
            return math.degrees(
                math.acos(
                    np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))))

        def _lattice_parameters(L):
            a = np.linalg.norm(L[:, 0])
            b = np.linalg.norm(L[:, 1])
            c = np.linalg.norm(L[:, 2])
            alpha = _angle(L[:, 1], L[:, 2])
            beta = _angle(L[:, 0], L[:, 2])
            gamma = _angle(L[:, 0], L[:, 1])
            return {
                'a': a,
                'b': b,
                'c': c,
                'alpha': alpha,
                'beta': beta,
                'gamma': gamma
            }

        self.lattice_parameters = _lattice_parameters(self.lattice_matrix)

        (stdout, stderr, returncode) = proj.capture("sym")

        # lattice symmetry
        self.lattice_symmetry_s = syminfo.lattice_symmetry(stdout)
        self.lattice_symmetry_hm = syminfo.hm_symmetry(self.lattice_symmetry_s)
        self.lattice_system = syminfo.lattice_system(self.lattice_symmetry_s)

        # crystal symmetry
        self.crystal_symmetry_s = syminfo.crystal_symmetry(stdout)
        self.crystal_symmetry_hm = syminfo.hm_symmetry(self.crystal_symmetry_s)
        self.crystal_system = syminfo.crystal_system(self.crystal_symmetry_s)
        self.crystal_family = syminfo.crystal_family(self.crystal_symmetry_s)
        self.space_group_number = syminfo.space_group_number_map[
            self.crystal_symmetry_s]

        # # composition (v0.2.X: elements and components are identical, only 'occupation' allowed)
        # with open(self.proj.dir.composition_axes()) as f:
        #     raw_composition_axes = json.load(f)
        #
        # print(raw_composition_axes)
        #
        # self.components = raw_composition_axes['possible_axes']['0'][
        #     'components']
        # self.elements = self.components
        # self.n_independent_compositions = raw_composition_axes[
        #     'possible_axes']['0']['independent_compositions']
        # self.degrees_of_freedom = ['occupation']


class CompositionAxes(object):
    """A composition axes object

    Attributes
    ----------

        name: str
          composition axes name

        components: List[str]
          occupational components

        n_independent_compositions: int
          number of independent composition axes

        mol_formula: str
          number of each component in terms of the parametric compositions

        param_formula: str
          parametric compositions in terms of the number of components

        end_members: dict of np.array of shape=(n_components,)
          the number of components per unit cell in each end member state, in form:
          {'origin':np.array, 'a':np.array, 'b', np.array, ...}. Order matches
          that given by self.components.


    """
    def __init__(self, name, data):
        self._name = name
        self._data = data

        self._end_members = {}
        for c in ascii_lowercase:
            if c in self._data:
                self.end_members[c] = np.array(self._data[c])[:, 0]
            else:
                break
        self._end_members['origin'] = np.array(self._data['origin'])[:, 0]

    @property
    def name(self):
        return self._name

    @property
    def components(self):
        return self._data['components']

    @property
    def n_independent_compositions(self):
        return self._data['independent_compositions']

    @property
    def mol_formula(self):
        return self._data['mol_formula']

    @property
    def param_formula(self):
        return self._data['param_formula']

    @property
    def origin(self):
        return self._origin

    @property
    def end_members(self):
        return self._end_members



"""Write without buffering"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *


def _print(*args, **kwargs):
    """Print and always flush"""
    sep = kwargs.get('sep', ' ')
    end = kwargs.get('end', '\n')
    file = kwargs.get('file', sys.stdout)
    print(*args, sep=sep, end=end, file=file)
    file.flush()



"""Functions that are Python2/3 compatible"""
import io
import six

# Notes on Python2/3 compatible io for casm:

# For pandas csv files:
# with open('test.csv', compat.pandas_wmode()) as f:
#     f.write('# ')  # will make this optional in a future version
#     df.to_csv(f, sep=compat.str(' '), index=False)
#
# with open('test.csv', compat.pandas_rmode()) as f:
#     if compat.peek(f) == '#':
#         f.read(1)
#     df = pandas.read_csv(f, sep=compat.str(' +'), engine='python')
#     print(df)

# For json files:
# with open('test.json', 'wb') as f:
#     f.write(six.u(json.dumps(data, indent=2)).encode('utf-8'))
#
# with open('test.json', 'rb') as f:
#     data = json.loads(f.read().decode('utf-8'))
#     print(data)

# For pickle files:
# with open('test.pkl', 'wb') as f:
#     pickle.dump(data, f, protocol=2)
#
# with open('test.pkl', 'rb') as f:
#     data = pickle.load(f)
#     print(data)


def str(val):
    """Convert to text to Python2/3 native string literal type"""
    if six.PY2:
        return val.encode('utf-8')
    else:
        return val


def peek(f, size=1):
    pos = f.tell()
    data = f.read(size)
    f.seek(pos)
    return data


def pandas_wmode():
    if six.PY2:
        return 'wb'
    else:
        return 'w'


def pandas_rmode():
    return 'r'


def native_io():
    """Return Python2/3 IO stream (BytesIO or StringIO)"""
    if six.PY2:
        return io.BytesIO()
    else:
        return io.StringIO()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import sys
import subprocess
import time


def main():
    print("Begin casm.view")

    if len(sys.argv) != 3:
        print(
            "Usage: casm.view '...command to call visualization software...' /path/to/POSCAR"
        )
        sys.exit()

    VESTAcommand = sys.argv[1]
    POSCARpath = sys.argv[2]

    print("Reading POSCAR:")
    print(open(POSCARpath).read())
    subprocess.Popen((VESTAcommand + " " + POSCARpath).split())

    time.sleep(1)


if __name__ == "__main__":
    main()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import time
from casm.learn.fit import make_fitting_data, make_estimator, make_selector, \
  add_individual_detail, print_halloffame
import casm.learn.model_selection
from casm.learn.evolve import GeneticAlgorithm, IndividualBestFirst, PopulationBestFirst


def fit_and_select(input,
                   save=True,
                   verbose=True,
                   read_existing=True,
                   hall=None):
    """
  Arguments
  ---------

    input: dict
      The input settings as a dict

    save: boolean, optional, default=True
      Save a pickle file containing the training data and scoring metric. The file
      name, which can be specified by input["fit_data_filename"], defaults to "fit_data.pkl".

    verbose: boolean, optional, default=True
      Print information to stdout.

    read_existing: boolean, optional, default=True
      If it exists, read the pickle file containing the training data and scoring
      metric. The file name, which can be specified by input["fit_data_filename"],
      defaults to "fit_data.pkl".

    hall: deap.tools.HallOfFame, optional, default=None
      A Hall Of Fame to add resulting individuals to


  Returns
  -------

    (fdata, estimator, selector)

    fdata: casm.learn.FittingData
      A FittingData instance containing the problem data.

    estimator:  estimator object implementing 'fit'
      The estimator specified by the input settings.

    selector:  selector object implementing 'fit' and having either a
               'get_support()' or 'get_halloffame()' member
      The feature selector specified by the input settings.


  """
    # construct FittingData
    fdata = make_fitting_data(input,
                              save=True,
                              verbose=verbose,
                              read_existing=True)

    # construct model used for fitting
    estimator = make_estimator(input, verbose=verbose)

    # feature selection
    selector = make_selector(input,
                             estimator,
                             scoring=fdata.scoring,
                             cv=fdata.cv,
                             penalty=fdata.penalty,
                             verbose=verbose)

    if not hasattr(selector, "get_halloffame") and not hasattr(
            selector, "get_support"):
        raise Exception(
            "Selector has neither 'get_halloffame' nor 'get_support'")

    # fit
    if verbose:
        print("# Fit and select...")
    t = time.perf_counter()
    selector.fit(fdata.weighted_X, fdata.weighted_y)

    # print calculation properties that may be of interest
    if verbose:
        # custom for SelectFromModel
        if hasattr(selector, "estimator_") and hasattr(selector.estimator_,
                                                       "n_iter_"):
            print("#   Iterations:", selector.estimator_.n_iter_)
        if hasattr(selector, "estimator_") and hasattr(selector.estimator_,
                                                       "alpha_"):
            print("#   Alpha:", selector.estimator_.alpha_)
        if hasattr(selector, "threshold"):
            print("#   Feature selection threshold:", selector.threshold)
        print("#   DONE  Runtime:", time.perf_counter() - t, "(s)\n")

    if hall is not None:
        # store results: Assume selector either has a 'get_halloffame()' attribute, or 'get_support()' member
        if hasattr(selector, "get_halloffame"):
            if verbose:
                print("Adding statistics...")
            selector_hall = selector.get_halloffame()
            for i in range(len(selector_hall)):
                add_individual_detail(selector_hall[i],
                                      estimator,
                                      fdata,
                                      input,
                                      selector=selector)
            if verbose:
                print("  DONE\n")

            if verbose:
                print("Result:")
            print_halloffame(selector_hall)

            hall.update(selector_hall)

        elif hasattr(selector, "get_support"):
            indiv = casm.learn.creator.Individual(selector.get_support())
            if verbose:
                print("Adding statistics...")
            indiv.fitness.values = casm.learn.model_selection.cross_val_score(
                estimator,
                fdata.weighted_X,
                indiv,
                y=fdata.weighted_y,
                scoring=fdata.scoring,
                cv=fdata.cv,
                penalty=fdata.penalty)
            add_individual_detail(indiv,
                                  estimator,
                                  fdata,
                                  input,
                                  selector=selector)
            if verbose:
                print("  DONE\n")

            if verbose:
                print("Result:")
            print_halloffame([indiv])

            hall.update([indiv])

    return (fdata, estimator, selector)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import numpy as np
import os, re
"""
Sample Species File:
POTCAR_DIR_PATH = /absolute/path/to/vasp_potentials
SPECIES    ALIAS    POTCAR  POTCAR_location    MAGMOM    NBANDS
Mn3        Mn       0       -                  3         5
Mn4        Mn       1       PAW_PBE/Mn         4         6

"""


class SpeciesError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class SpeciesDict(dict):
    """
        SpeciesDict subclasses dict so it can hold additional metadata without disrupting routines that
            rely on species_setting containing *only* key/value pairs corresponding to IndividualSpecies
    """
    def set_available_tags(self, tags):
        """ Set a metadata member that lists the available tags listed in a SPECIES file """
        self.tags = tags


class IndividualSpecies:
    """
        The IndividualSpecies class contains:
            self.name: the name as listed in the POS file
            self.alias: the species file lists the name, POSCAR printing uses this string
            self.tags: (dict) the tags that need to be modified in the INCAR for this specie (ex. MAGMOM = -1)
                        All values are stored as strings.
            self.write_potcar: decides if a POTCAR file needs to be written for this specie
                     (for example, you have Mn3 and Mn4 but want to write only one POTCAR, set
                      one of them to true and the other to false)
            self.potcardir_base: common directory for all POTCARs
            self.potcar_location: location of POTCAR directory relative to self.potcardir_base
            self.potcardir: directory containing particular POTCAR (self.potcardir_base joined with self.potcar_location)
    """
    def __init__(self, values, tags, potcardir_base):
        """ Construct an IndividualSpecies.

            Args:
                values: (str list) entries in SPECIES file row
                tags: (str list) column names 4+ in SPECIES file, INCAR tags that need to be modified
                potcardir_base: (str) common directory for all POTCARs
        """
        if len(values) != (len(tags) + 4):
            raise SpeciesError(
                "Length of values != length of tags + 4.\nvalues = " +
                str(values) + "\ntags = " + str(tags))
        self.name = values[0]
        self.alias = values[1]
        try:
            self.write_potcar = not (int(values[2]) == 0)
        except ValueError:
            raise SpeciesError("Could not read POTCAR: " + str(values))
        self.potcardir_base = potcardir_base
        self.potcar_location = values[3]
        self.potcardir = os.path.join(self.potcardir_base,
                                      self.potcar_location.lstrip("/"))
        self.tags = dict()
        for i, key in enumerate(tags):
            self.tags[key] = values[i + 4]

    def write_header(self, file):
        """ Write header to a file """
        file.write("POTCAR_DIR_PATH = " + self.potcardir_base)
        headers = "{0:<12} {1:<12} {2:<12} {3:<36} ".format(
            "SPECIES", "ALIAS", "POTCAR", "POTCAR_location")
        for key in sorted(self.tags.keys()):
            headers += "{0:<12}".format(key)
        file.write(headers)

    def write(self, file):
        """ Write IndividualSpecies line"""
        values = "{0:<12} {1:<12} {2:<12} {3:<36} ".format(
            self.name, self.alias, self.write_potcar, self.potcar_location)
        for key in sorted(self.tags.keys()):
            values += "{0:<12}".format(self.tags[key])
        file.write(values)

    def print_all(self):
        print(self.name)
        print(self.alias)
        print(self.tags)
        print(self.write_potcar)
        print(self.potcardir_base)
        print(self.potcar_location)
        print(self.potcardir)


# replacing Species class with 'species_settings' function that returns a dict of IndividualSpecies

#class Species:
#    """
#    The Species class:
#        -specieList: a list of IndividualSpecies stored as key value pairs
#                     (specieList['Mn3'] would be the Mn3 specie)
#    """
#
#    def __init__(self,filename):
#        self.POTCAR_DIR_PATH = None
#        self.specieList = []
#        self.read(filename)
#
#    def read(self,filename):
#        try:
#            file = open(filename)
#        except IOError:
#            raise SpecieError('IOError',filename)
#
#        # Read POTCAR_DIR_PATH from first line
#        firstline = file.readline().strip().split()
#        if len(firstline) != 3 or firstline[0] != "POTCAR_DIR_PATH" or firstline[1] != "=" or not os.path.isdir(firstline[2].strip("\'\"")):
#            raise SpecieError(firstline,"expected: POTCAR_DIR_PATH = /path/to/POTCAR_DIR")
#        self.POTCAR_DIR_PATH = firstline[2].strip("\'\"")
#
#        # Parsing the header
#        header = file.readline().strip()
#        columnNames = header.split()
#        if len(columnNames)<4:
#            raise SpecieError(header,'insufficient number of columns')
#        tagList = columnNames[4:]
#        self.specieList = []
#        for line in file:
#            self.specieList.append(IndividualSpecie(line.strip(),tagList))
#
#
#    def print_all(self):
#        for specie in self.specieList:
#            specie.print_all()


def species_settings(filename):
    """ Returns a SpeciesDict of IndividualSpecies objects, with keys equal to their names. """
    try:
        file = open(filename)
    except IOError:
        raise SpeciesError("Could not open: '" + filename + "'")

    # Read POTCAR_DIR_PATH from first line
    line = file.readline()
    m = re.match("POTCAR_DIR_PATH\s*=\s*(.*)", line)
    if not m:
        raise SpeciesError(
            "Could not read POTCAR_DIR_PATH.\nExpected: POTCAR_DIR_PATH = /path/to/POTCAR_DIR\nFound: '"
            + line + "'")
    POTCAR_DIR_PATH = m.group(1)

    # Parsing the header
    header = file.readline().strip()
    column_names = header.split()
    if len(column_names) < 4:
        raise SpeciesError("Insufficient number of columns in SPECIES file")
    tags = column_names[4:]
    species_settings = SpeciesDict()
    species_settings.set_available_tags(tags)
    for line in file:
        if line.strip():
            values = line.strip().split()
            species_settings[values[0]] = IndividualSpecies(
                values, tags, POTCAR_DIR_PATH)

    file.close()

    return species_settings


def write_species_settings(species, filename):
    """ Write a SPECIES file from a species dict """
    try:
        file = open(filename, 'w')
    except:
        raise SpeciesError("Could not open file for writing: '" + filename +
                           "'")
    species.keys()[0].write_header(file)
    for s in sorted(species.keys()):
        species[s].write(file)
    file.close()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import re
from casm.wrapper.misc import remove_chars
from casm.vasp.io import attribute_classes

# List of tags in VASP sorted by the data type associated with it
VASP_TAG_INT_LIST = [
    'ialgo', 'ibrion', 'icharg', 'images', 'ismear', 'ispin', 'istart', 'isym',
    'lorbit', 'nbands', 'ndav', 'ngx', 'ngxf', 'ngy', 'ngyf', 'ngz', 'ngzf',
    'npar', 'ncore', 'spind', 'nsw', 'isif', 'kpar', 'voskown', 'nsim',
    'nedos', 'lmaxfock', 'lmaxmix', 'nkred', 'ivdw', 'nelmin', 'nelm',
    'nelmdl', 'ldautype', 'ldauprint', 'ldauprint', 'ichain'
]
VASP_TAG_FLOAT_LIST = [
    'ediff', 'ediffg', 'emax', 'emin', 'encut', 'potim', 'sigma', 'enmax',
    'symprec', 'time', 'hfscreen', 'amix', 'bmix', 'amix_mag', 'bmix_mag',
    'spring'
]
VASP_TAG_BOOL_LIST = [
    'lcharg', 'lsorbit', 'lwave', 'lscalapack', 'lscalu', 'lplane', 'lhfcalc',
    'shiftred', 'evenonly', 'oddonly', 'addgrid', 'ldau', 'lasph', 'lclimb',
    'ldneb', 'lnebcell', 'ltangentold'
]
# Site-wise list of arrays of FLOAT
VASP_TAG_SITEF_LIST = ['magmom', 'rwigs']
# Species-wise list of arrays of FLOAT
VASP_TAG_SPECF_LIST = ['ldauu', 'ldauj']
# Site-wise list of arrays of INT
VASP_TAG_SPECI_LIST = ['ldaul']

VASP_TAG_STRING_LIST = [
    'algo', 'prec', 'system', 'precfock', 'lreal', 'metagga'
]

# The master list of VASP tags is a union of the above -> need to allow for 'miscellaneous' ?
VASP_TAG_LIST = VASP_TAG_INT_LIST + VASP_TAG_SITEF_LIST + VASP_TAG_SPECI_LIST + \
    VASP_TAG_BOOL_LIST + VASP_TAG_FLOAT_LIST + \
    VASP_TAG_STRING_LIST + VASP_TAG_SPECF_LIST


class IncarError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class Incar(object):
    """
    The INCAR class contains:
        tags: a dict of all INCAR settings

    All input tags and associated values are stored as key-value pairs in the dicionary called 'tags'.
   """
    def __init__(self,
                 filename,
                 species=None,
                 poscar=None,
                 sort=True,
                 dof_info=None):
        """ Construct an Incar object from 'filename'"""
        self.read(filename, species, poscar, sort, dof_info)

    def read(self,
             filename,
             species=None,
             poscar=None,
             sort=True,
             dof_info=None):
        """ Read an INCAR file """
        self.tags = dict()
        try:
            file = open(filename, 'r')
        except:
            raise IncarError("Could not open file: '" + filename + "'")

        # parse INCAR into self.tags
        for line in file:
            line = re.split('=', re.split('#', line)[0])
            if len(line) == 2:
                self.tags[line[0].strip()] = line[1].strip()
        self._verify_tags()
        self._make_natural_type()

        if species != None:
            self.update(species, poscar, sort, dof_info)

        file.close()

    def _make_natural_type(self):
        """ Convert self.tags values from strings into their 'natural type' (int, float, etc.) """
        for tag in self.tags:
            if self.tags[tag] is None or str(self.tags[tag]).strip() == "":
                self.tags[tag] = None
            else:
                if tag.lower() in VASP_TAG_INT_LIST:
                    try:
                        self.tags[tag] = int(self.tags[tag])
                    except ValueError:
                        raise IncarError("Could not convert '" + tag +
                                         "' : '" + self.tags[tag] + "' to int")
                elif tag.lower() in VASP_TAG_FLOAT_LIST:
                    try:
                        self.tags[tag] = float(self.tags[tag].lower().replace(
                            'd', 'e'))
                    except ValueError:
                        raise IncarError("Could not convert '" + tag +
                                         "' : '" + self.tags[tag] +
                                         "' to float")
                elif tag.lower() in VASP_TAG_BOOL_LIST:
                    if not self.tags[tag].lower() in ['.true.', '.false.']:
                        raise IncarError("Could not find '" + tag + "' : '" +
                                         self.tags[tag].lower() +
                                         "' in ['.true.','.false.']")
                    else:
                        self.tags[tag] = (self.tags[tag].lower() == '.true.')
                elif tag.lower() in VASP_TAG_SITEF_LIST + VASP_TAG_SPECF_LIST:
                    temp = []
                    for value in self.tags[tag].split():
                        try:
                            item = value.split('*')
                            if len(item) == 1:
                                temp.append(float(value))
                            else:
                                if item[0] != 0:
                                    temp.append(
                                        str(item[0]) + '*' +
                                        str(float(item[1])))
                        except ValueError:
                            raise IncarError("Could not convert '" + tag +
                                             "' : '" + self.tags[tag] +
                                             "' to float list")
                    self.tags[tag] = temp
                elif tag.lower() in VASP_TAG_SPECI_LIST:
                    temp = []
                    for value in self.tags[tag].split():
                        try:
                            temp.append(int(value))
                        except ValueError:
                            raise IncarError("Could not convert '" + tag +
                                             "' : '" + self.tags[tag] +
                                             "' to int list")
                    self.tags[tag] = temp
                elif tag.lower() in VASP_TAG_STRING_LIST:
                    self._check_string_tag(tag, self.tags[tag])

    def _check_string_tag(self, tag, value):
        """ Check that string-valued tags are allowed values """
        if tag.lower() == 'prec':
            if value.lower() not in [
                    'low', 'medium', 'high', 'normal', 'single', 'accurate'
            ]:
                raise IncarError("Unknown 'prec' value: '" + value)
        elif tag.lower() == 'algo':
            if value.lower() not in [
                    'normal', 'veryfast', 'fast', 'conjugate', 'all', 'damped',
                    'subrot', 'eigenval', 'none', 'nothing', 'chi', 'gw0',
                    'gw', 'scgw0', 'scgw'
            ]:
                raise IncarError("Unknown 'algo' value: '" + value)

    def _verify_tags(self):
        """ Check that only allowed INCAR tags are in self.tags """
        for tag in self.tags:
            if tag.lower() in VASP_TAG_LIST:
                continue
            else:
                print(("Warning: unknown INCAR tag '" + tag +
                       "' with value '" + str(self.tags[tag]) + "'"))

    def update(self, species, poscar, sort=True, dof_info=None):
        """ Update Incar object to reflect Species settings """

        #TODO: Need a way to deal with conflicting dof setups. There won't be an issue if user isn't stupid.
        #TODO: Need a better update function, it's very construed and difficult to understand. Apologies!

        if dof_info is not None and dof_info.atom_dofs is not None:
            if "Cmagspin" in list(dof_info.atom_dofs.keys()):
                vasp_input_tags_to_append = attribute_classes.CmagspinAttr(
                    dof_info).vasp_input_tags()
                self.tags.update(vasp_input_tags_to_append)

        if sort == False:
            # for each 'tag' in the IndividualSpecies, create a list in self.tags
            for key in list(species.values())[0].tags.keys():
                if key.lower() in (VASP_TAG_INT_LIST + VASP_TAG_FLOAT_LIST):
                    self.tags[key] = 0.
                    for site in poscar.basis:
                        self.tags[key] += float(
                            species[site.occupant].tags[key])
                    if key.lower() in VASP_TAG_INT_LIST:
                        self.tags[key] = int(self.tags[key])
                else:
                    self.tags[key] = []
                    if key.lower() in (VASP_TAG_SPECF_LIST +
                                       VASP_TAG_SPECI_LIST):
                        # add the value of the 'tag' for each species into the self.tags list
                        for spec in poscar.type_atoms_alias:
                            self.tags[key].append(species[spec].tags[key])
                    else:
                        # add the value of the 'tag' for each atom into the self.tags list
                        for site in poscar.basis:
                            self.tags[key].append(
                                species[site.occupant].tags[key])
        else:
            pos = poscar.basis_dict()
            # for each 'tag' in the IndividualSpecies, create a list in self.tags
            for key in list(species.values())[0].tags.keys():
                # for key in species[species.keys()[0]].tags.keys():
                if key.lower() in (VASP_TAG_INT_LIST + VASP_TAG_FLOAT_LIST):
                    self.tags[key] = 0.
                    for site in poscar.basis:
                        self.tags[key] += float(
                            species[site.occupant].tags[key])
                    if key.lower() in VASP_TAG_INT_LIST:
                        self.tags[key] = int(self.tags[key])
                else:
                    self.tags[key] = []
                    # add the value of the 'tag' for each atom into the self.tags list
                    for alias in sorted(pos.keys()):
                        if key.lower() in (VASP_TAG_SPECF_LIST +
                                           VASP_TAG_SPECI_LIST):
                            # for species-specific tags, use the value specified for the
                            # species whose pseudopotential is being used for this alias
                            for name in species.keys():
                                if species[name].alias == alias and species[
                                        name].write_potcar:
                                    self.tags[key].append(
                                        species[name].tags[key])
                                    break
                        else:
                            for name in species.keys():
                                count = 0
                                for site in pos[alias]:
                                    if site.occupant == name:
                                        count += 1
                                if species[name].alias == alias:
                                    if count > 0:
                                        self.tags[key].append(
                                            str(count) + "*" +
                                            str(species[name].tags[key]))

    def write(self, filename):
        try:
            incar_write = open(filename, 'w')
        except IOError as e:
            raise e
        for tag in self.tags:
            if self.tags[tag] is None or str(self.tags[tag]).strip() == "":
                pass
            else:
                if tag.lower() in VASP_TAG_SITEF_LIST + VASP_TAG_SPECF_LIST:
                    incar_write.write('{} = {}\n'.format(
                        tag.upper(), remove_chars(self.tags[tag], "[\[\],']")))
                elif tag.lower() in VASP_TAG_SPECI_LIST:
                    incar_write.write('{} = {}\n'.format(
                        tag.upper(), remove_chars(self.tags[tag], "[\[\],']")))
                elif tag.lower() in VASP_TAG_BOOL_LIST:
                    if self.tags[tag] == True:
                        incar_write.write('{} = .TRUE.\n'.format(tag.upper()))
                    else:
                        incar_write.write('{} = .FALSE.\n'.format(tag.upper()))
                else:
                    incar_write.write('{} = {}\n'.format(
                        tag.upper(), self.tags[tag]))
        incar_write.close()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import argparse
import deap.tools
import json
import os
import six
import sys

import casm.learn
from casm.project import Project, Selection, write_eci


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(
        description='Fit cluster expansion coefficients (ECI)')
    parser.add_argument('--desc',
                        help='Print extended usage description',
                        action="store_true")
    parser.add_argument('-s',
                        '--settings',
                        nargs=1,
                        help='Settings input filename',
                        type=str)
    parser.add_argument(
        '--format',
        help=
        'Hall of fame print format. Options are "details", "json", or "csv".',
        type=str,
        default=None)
    #parser.add_argument('--path', help='Path to CASM project. Default assumes the current directory is in the CASM project.', type=str, default=os.getcwd())
    parser.add_argument('--settings-format',
                        help='Print input file description',
                        action="store_true")
    parser.add_argument('--exLasso',
                        help='Print example input file using Lasso',
                        action="store_true")
    parser.add_argument('--exLassoCV',
                        help='Print example input file using LassoCV',
                        action="store_true")
    parser.add_argument(
        '--exRFE',
        help=
        'Print example input file using Recursive Feature Elimination (RFE)',
        action="store_true")
    parser.add_argument('--exGeneticAlgorithm',
                        help='Print example input file using GeneticAlgorithm',
                        action="store_true")
    parser.add_argument(
        '--exIndividualBestFirst',
        help='Print example input file using IndividualBestFirst',
        action="store_true")
    parser.add_argument(
        '--exPopulationBestFirst',
        help='Print example input file using PopulationBestFirst',
        action="store_true")
    parser.add_argument('--exDirectSelection',
                        help='Print example input file using DirectSelection',
                        action="store_true")
    parser.add_argument('--hall',
                        help='Print hall of fame summary',
                        action="store_true")
    parser.add_argument(
        '--indiv',
        nargs='+',
        help='Specify particular individuals by index in hall of fame',
        type=int)
    parser.add_argument('--select',
                        nargs=1,
                        help='Select individual to use',
                        type=int)
    parser.add_argument(
        '--checkhull',
        help='Check convex hull properties using the provided selection',
        action="store_true",
        default=False)
    parser.add_argument(
        '--checkspecs',
        help='Output data and cv files containing the current problem specs',
        action="store_true",
        default=False)
    parser.add_argument('-q',
                        '--quiet',
                        help='Quiet output',
                        action="store_true",
                        default=False)
    args = parser.parse_args(argv)

    args.verbose = not args.quiet

    if args.settings_format:
        casm.learn.print_input_help()
        return

    if args.exLasso:
        print(six.u(json.dumps(casm.learn.example_input_Lasso(), indent=2)))
        return
    elif args.exLassoCV:
        print(six.u(json.dumps(casm.learn.example_input_LassoCV(), indent=2)))
        return
    elif args.exRFE:
        print(six.u(json.dumps(casm.learn.example_input_RFE(), indent=2)))
        return
    elif args.exGeneticAlgorithm:
        print(
            six.u(
                json.dumps(casm.learn.example_input_GeneticAlgorithm(),
                           indent=2)))
        return
    elif args.exIndividualBestFirst:
        print(
            six.u(
                json.dumps(casm.learn.example_input_IndividualBestFirst(),
                           indent=2)))
        return
    elif args.exPopulationBestFirst:
        print(
            six.u(
                json.dumps(casm.learn.example_input_PopulationBestFirst(),
                           indent=2)))
        return
    elif args.exDirectSelection:
        print(
            six.u(
                json.dumps(casm.learn.example_input_DirectSelection(),
                           indent=2,
                           cls=casm.NoIndentEncoder)))
        return

    if args.settings:

        if args.verbose:
            print("Loading", args.settings[0])

        input = casm.learn.open_input(args.settings[0])

        if args.hall:

            halloffame_filename = input["halloffame_filename"]
            # print Hall of Fame summary
            existing_hall = casm.learn.open_halloffame(halloffame_filename,
                                                       args.verbose)

            if args.indiv:
                casm.learn.print_individual(existing_hall,
                                            args.indiv,
                                            format=args.format)
            elif args.hall:
                casm.learn.print_halloffame(existing_hall, format=args.format)

        elif args.checkhull:

            halloffame_filename = input["halloffame_filename"]
            # print Hall of Fame summary
            existing_hall = casm.learn.open_halloffame(halloffame_filename,
                                                       args.verbose)

            casm.learn.fit.checkhull(input,
                                     existing_hall,
                                     indices=args.indiv,
                                     verbose=args.verbose)

            # pickle hall of fame
            casm.learn.save_halloffame(existing_hall, halloffame_filename,
                                       args.verbose)

        elif args.select:

            specs = input["problem_specs"]
            halloffame_filename = input["halloffame_filename"]
            # print Hall of Fame summary
            existing_hall = casm.learn.open_halloffame(halloffame_filename,
                                                       args.verbose)

            proj = Project(specs["data"]["kwargs"]["project_path"])

            index = args.select[0]
            indiv = existing_hall[index]

            write_eci(proj,
                      indiv.eci,
                      casm.learn.to_json(index, indiv),
                      verbose=args.verbose)

        elif args.checkspecs:

            casm.learn.checkspecs(input, verbose=args.verbose)

        else:

            # construct hall of fame
            specs = input["problem_specs"]
            halloffame_filename = input["halloffame_filename"]
            halloffame_size = input["n_halloffame"]
            hall = casm.learn.create_halloffame(halloffame_size)
            if args.verbose:
                print("# Hall of Fame size:", halloffame_size, "\n")

            if os.path.exists(halloffame_filename):
                existing_hall = casm.learn.open_halloffame(
                    halloffame_filename, args.verbose)
                hall.update(existing_hall)

            # run fitting
            if input["feature_selection"]["method"] == "DirectSelection":
                casm.learn.direct_fit(input, verbose=args.verbose, hall=hall)
            else:
                casm.learn.fit_and_select(input,
                                          verbose=args.verbose,
                                          hall=hall)

            # pickle hall of fame
            casm.learn.save_halloffame(hall, halloffame_filename, args.verbose)

    elif args.desc:

        print("""
    
    1) Specify the problem:
      
      'casm-learn' helps solve the problem:
      
        X*b = y,
      
      where:
        
        X: 2d matrix of shape (n_samples, n_features)
          The correlation matrix, holding the evaluated basis functions. The 
          entry X[config, bfunc] holds the average value of the 'bfunc' cluster
          basis function for configuration 'config'. The number of configurations
          is 'n_samples' and the number of cluster basis functions is 'n_features'.
        
        y: 1d matrix of shape (n_samples, 1)
          The calculated properties being fit to. The most common case is that
          y[config] holds the formation energy calculated for configuration 
          'config'.
        
        b: 1d matrix of shape (n_features, 1)
          The effective cluster interactions (ECI) being solved for.
      
      To specify this problem, the 'casm-learn' input file specifies which
      configurations to fit to (the training data), how to weight the 
      configurations, and how to compare solutions via cross-validation. 
            
      
      Training data may be input via a 'casm select' output file. The default
      name expected is 'train'. So to use all calculated configurations, you 
      could create a directory in your CASM project where you will perform
      fitting and generate a 'train' file:
        
        cd /my/casm/project
        mkdir fit_1 && cd fit_1
        casm select --set is_calculated -o train
      
      
      Example 'casm-learn' JSON input files can be output by the
      'casm-learn --exMethodName' options:
        
        casm-learn --exGeneticAlgorithm > fit_1_ga.json
        casm-learn --exRFE > fit_1_rfe.json
        ...etc..
      
      By default, these settings files are prepared for fitting formation_energy,
      using the 'train' configuration selection. Edit the file as needed, and
      see 'casm-learn --settings-format' for help.
      
      
      When weighting configurations, the problem is transformed:
      
        X*b = y  ->  L*X*b = L*y, 
  
      where, W = L*L.tranpose(): 
        
        W: 2d matrix of shape (n_samples, n_samples)
          The weight matrix is specified in the casm-learn input file. If the 
          weighting method provides 1-dimensional input (this is typical, i.e.
          a weight for each configuration), in an array called 'w', then:
          
            W = diag(w)*n_samples/sum(w),
            
          diag(w) being the diagonal matrix with 'w' along the diagonal.
      
      
      A cross-validation score is used for comparing generated ECI. The cv score
      reported is:
  
        cv = sqrt(mean(scores)) + N_nonzero_eci*penalty, 
  
      where:
      
        scores: 1d array of shape (number of train/test sets)
          The mean squared error calculated for each training/testing set
        
        N_nonzero_eci: int
          The number of basis functions with non-zero ECI
          
        penalty: number, optional, default=0.0
          Is the user-input penalty per basis function that can be used to
          favor solutions with a small number of non-zero ECI
      
      See 'casm-learn --settings-format' for help specifying the cross-validation
      training and test sets using options from scikit-learn. It is usually
      important to use the 'shuffle'=true option so that configurations are
      randomly added to train/test sets and not ordered by supercell size.
      
      
      When you run 'casm-learn' with a new problem specification the first time, 
      it generates a "problem specs" file that stores the training data, weights, 
      and cross-validation train/test sets. Then, when running subsequent times,
      the data can be loaded more quickly, and the cross-validation can be 
      performed using the same train/test sets. 'casm-learn' will attempt to
      prevent you from re-running with a different problem specification so that
      solutions can be compared via their cv score in an "apples-to-apples" 
      manner. The default name for the "specs" file is determined from the input
      filename. For example, 'my_input_specs.pkl' is used if the input file is 
      named 'my_input.json'. See 'casm-learn --settings-format' for more help.
      
      
      The '--checkspecs' option can be used to write output files with the 
      generated problem specs data. Amont other things, this can be used to 
      adjust weights manually or save and re-use train/test sets. See 
      'casm-learn --settings-format' for more help.
      
    
    2) Select estimator and feature selection methods
      
      The "estimator" option specifies a linear model estimator that determines
      how to solve the linear problem L*X*b = L*b, for b.
      
      The "feature_selection" option specifies a feature selection method that
      determines which features (ECI) should be considered for the solution. The
      remaining are effectively set to 0.0 when calculating the cluster 
      expansion. Generally there is a tradeoff: By limiting the number of 
      features included in the cluster expansion Monte Carlo calculations can be
      more efficient, but at a possible loss of accuracy. Be careful to avoid
      overfitting however. If your cross validation scheme does not provide 
      enough testing data, you may fit your training data very well, but not
      have an accurate extrapolation to other configurations.
      
      See 'casm-learn --settings-format' for help specifying the estimator and
      feature selection methods. Assuming you are using the GeneticAlgorithm and
      have named your input file 'fit_1_ga.json', run:
        
        casm-learn -s fit_1_ga.json
      
      'casm-learn' will run and eventually store its results.  For a single 
      problem specification (step 1, the settings in "problem_specs"
      in the 'casm-learn' input file), you may try many different estimation
      and feature selection methods and use the cv score to compare results. All
      the results for a single problem specification can be stored in a 'Hall Of
      Fame' that collects the N individual solutions with the best cv scores. To
      view these results use:
        
        casm-learn -s fit_1_ga.json --hall
      
      For more details, or to output the results for further analysis in JSON or 
      CSV format, there is a '--format' option. To view only particular 
      individuals in the hall of fame, there is a '--indiv' option.
    
    
    3) Analyze results
      
      The above steps (1) and (2) may be repeated many times as you attempt to
      optimize your ECI. Solutions for different problems (i.e. different 
      weighting schemes, re-calculating with more training data) may be compared 
      based on scientific knowledge, for instance, which predicts the 0K ground 
      state configurations correctly, or from analysis of Monte Carlo results.
      
      The '--checkhull' option provides a simple way to check the 0K ground 
      states and can create 'casm select' style output files with enumerated but
      uncalculated configurations that are predicted to be low energy. These can
      then be used to generate more training data and re-fit the ECI.
      
      When you have generated ECI that you wish to use in Monte Carlo 
      calculations, use the '--select' option to write an 'eci.json' file into
      your CASM project for the currently selected cluster expansion (as listed
      by 'casm settings -l).
    
    
    4) Use results
    
      Once an 'eci.json' file has been written, you can run Monte Carlo 
      calculations. See 'casm monte -h' and 'casm format --monte' for help. 
      
    """)

    else:

        parser.print_help()


if __name__ == "__main__":
    main()



"""CASM machine learning tools"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

## This part needs to be in global scope for parallization #####################
from deap import creator
from deap import base
from deap.tools import HallOfFame

# we'll want to minimize a cv score
creator.create("FitnessMin", base.Fitness, weights=(-1.0, ))

# each individual is a list of True or False indicating if each basis function should
# be included in the model
creator.create("Individual", list, fitness=creator.FitnessMin, input=None)


def empty_individual(n_features):
    selected = [False] * n_features
    indiv = creator.Individual(selected)
    return indiv


class EqualIndividual(object):
    """ 
  Functor class for checking equivalence of individuals in a HallOfFame.
    
  Attributes
  ----------
    
    rel_tol: number
      Individuals are considered equal if features selected are equal and 
      abs((A.fitness.values[0] - B.fitness.values[0]) / A.fitness.values[0] ) < rel_tol
    
  """
    def __init__(self, rel_tol=1e-6):
        """
    Arguments
    ---------
    
      rel_tol: number, optional, default=1e-6
        Individuals are considered equal if features selected are equal and 
        abs((A.fitness.values[0] - B.fitness.values[0]) / A.fitness.values[0] ) < rel_tol
    """
        self.rel_tol = rel_tol

    def __call__(self, A, B):
        """
    Arguments
    ---------
    
      A, B: Individuals to be compared
        Individuals are considered equal if features selected are equal and 
        abs((A.fitness.values[0] - B.fitness.values[0]) / A.fitness.values[0] ) < rel_tol
    
    Returns
    ---------
      result: boolean
        True if A and B compare equal, False otherwise
    
    """
        if A == B:
            if abs((A.fitness.values[0] - B.fitness.values[0]) /
                   A.fitness.values[0]) < self.rel_tol:
                return True
        return False


def create_halloffame(maxsize, rel_tol=1e-6):
    return HallOfFame(maxsize, similar=EqualIndividual(rel_tol))


################################################################################


from casm.learn.fit import example_input_Lasso, example_input_LassoCV, example_input_RFE, \
 example_input_GeneticAlgorithm, example_input_IndividualBestFirst, \
 example_input_PopulationBestFirst, example_input_DirectSelection, \
 open_input, set_input_defaults, \
 FittingData, TrainingData, \
 print_input_help, print_individual, print_population, print_halloffame, print_eci, \
 to_json, open_halloffame, save_halloffame, \
 checkspecs, checkhull

from casm.learn.feature_selection import fit_and_select
from casm.learn.direct_selection import direct_fit

__all__ = __all__ = [
    'create_halloffame', 'EqualIndividual', 'empty_individual',
    'example_input_Lasso', 'example_input_LassoCV', 'example_input_RFE',
    'example_input_GeneticAlgorithm', 'example_input_IndividualBestFirst',
    'example_input_PopulationBestFirst', 'example_input_DirectSelection',
    'open_input', 'set_input_defaults', 'FittingData', 'TrainingData',
    'print_input_help', 'print_individual', 'print_population',
    'print_halloffame', 'print_eci', 'to_json', 'open_halloffame',
    'save_halloffame', 'checkspecs', 'checkhull', 'fit_and_select',
    'direct_fit'
]



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

from casm.vasp.io.orbital_occupation import OrbitalOccupation

import os
import re
import gzip


class OutcarError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


# TODO update documentation here
class Outcar(object):
    """Parse OUTCAR files.

       Currently, just contains:
           self.complete = True/False
           self.slowest_loop = float
           self.kpoints = list of int, or none
           self.orbital_occupations = dict of OrbitalOccupation objects, from last step (only available when LDAUPRINT = 1 or 2). Keys are indices of only those sites for which occupation was printed.
    """
    def __init__(self, filename):
        self.filename = filename
        self.complete = False
        self.slowest_loop = None
        self.kpts = None
        self.lorbit = 0
        self.ispin = 1
        self.mag = None
        self.ngx = None
        self.ngy = None
        self.ngz = None
        self.found_ngx = False
        self.forces = []
        self.orbital_occupations = None

        self.read()

    def read(self):
        """Parse OUTCAR file.  Currently checks for:
                completion
                loop time
                kpoints
                lorbit (LORBIT value from INCAR)
                ispin (ISPIN value from INCAR)
                magnetization (if LORBIT = 1, 2, 11, 12)
                orbital_occupations
        """
        self.kpts = None
        if os.path.isfile(self.filename):
            if self.filename.split(".")[-1].lower() == "gz":
                f = gzip.open(self.filename)
            else:
                f = open(self.filename)
        elif os.path.isfile(self.filename + ".gz"):
            f = gzip.open(self.filename + ".gz")
        else:
            raise OutcarError("file not found: " + self.filename)

        force_index = False
        for line in f:
            try:
                if re.search("generate k-points for:", line):
                    self.kpts = list(map(int, line.split()[-3:]))
            except:
                pass

            try:
                if re.search("Total CPU time used", line):
                    self.complete = True
            except:
                raise OutcarError(
                    "Error reading 'Total CPU time used' from line: '" + line +
                    "'\nIn file: '" + self.filename + "'")

            try:
                if re.search("LOOP", line):
                    t = float(line.split()[-1])
                    if self.slowest_loop is None:
                        self.slowest_loop = t
                    elif t > self.slowest_loop:
                        self.slowest_loop = t
            except:
                pass

            try:
                if re.search("LORBIT", line):
                    self.lorbit = int(line.split()[2])
            except:
                pass

            try:
                if re.search("ISPIN", line):
                    self.ispin = int(line.split()[2])
            except:
                pass

            try:
                if re.search(r"magnetization \(x\)", line):
                    self.mag = []
                    for line in f:
                        try:
                            if re.match("[0-9]+", line.split()[0]):
                                self.mag.append(float(line.split()[-1]))
                            if re.match("tot", line.split()[0]):
                                break
                        except:
                            pass
            except:
                pass

            try:
                r = re.match(
                    r"\s*dimension x,y,z\s*NGX\s*=\s*([0-9]*)\s*NGY\s*=\s*([0-9]*)\s*NGZ\s*=\s*([0-9]*)\s*",
                    line)
                # if r:
                if r and not self.found_ngx:
                    self.ngx = int(r.group(1))
                    self.ngy = int(r.group(2))
                    self.ngz = int(r.group(3))
                    self.found_ngx = True
            except:
                pass

            if force_index:
                if '--' in line and len(self.forces) > 0:
                    force_index = False
                elif '--' not in line:
                    self.forces.append(list(map(float, line.split()[-3:])))

            try:
                if re.search("TOTAL-FORCE", line):
                    force_index = True
            except:
                pass

            # TODO: will this work for single-spin channel?
            try:
                if re.search("atom = *[0-9]+ *type = * [0-9]+  *l = *[0-9]+",
                             line):
                    if not self.orbital_occupations:
                        self.orbital_occupations = dict()
                    i = int(line.split()[2])
                    l = int(line.split()[8])
                    occupation_matrix_a = []
                    occupation_matrix_b = []
                    for inner_line in f:
                        try:
                            s = [float(x) for x in inner_line.split()]
                            if len(s) == 2 * l + 1:
                                if len(occupation_matrix_a) < 2 * l + 1:
                                    occupation_matrix_a.append(s)
                                else:
                                    occupation_matrix_b.append(s)
                        except:
                            pass

                        if len(occupation_matrix_b) == 2 * l + 1:
                            break
                    self.orbital_occupations[i - 1] = OrbitalOccupation(
                        occupation_matrix_a, occupation_matrix_b)
            except:
                pass

        f.close()



import json
import six
import uuid

# ---------------------------------------------------
# Some code to keep parts of a json string from being spread across multiple lines
# code from: http://stackoverflow.com/questions/13249415/can-i-implement-custom-indentation-for-pretty-printing-in-python-s-json-module
# answer: http://stackoverflow.com/a/25935321
# code by: http://stackoverflow.com/users/247623/erik-allik


class NoIndent(object):
    def __init__(self, value):
        self.value = value


class NoIndentEncoder(json.JSONEncoder):
    def __init__(self, *args, **kwargs):
        super(NoIndentEncoder, self).__init__(*args, **kwargs)
        self.kwargs = dict(kwargs)
        del self.kwargs['indent']
        self._replacement_map = {}

    def default(self, o):
        if isinstance(o, NoIndent):
            key = uuid.uuid4().hex
            self._replacement_map[key] = json.dumps(o.value, **self.kwargs)
            return "@@%s@@" % (key, )
        else:
            return super(NoIndentEncoder, self).default(o)

    def encode(self, o):
        result = super(NoIndentEncoder, self).encode(o)
        for k, v in six.iteritems(self._replacement_map):
            result = result.replace('"@@%s@@"' % (k, ), v)
        return result


# ---------------------------------------------------



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import sklearn.model_selection
import numpy as np
import pickle
from math import sqrt
from casm.learn.linear_model import LinearRegressionForLOOCV
from casm.learn.tools import indices


def LeaveOneOutForLLS(n_samples):
    """ 
  Returns train and test sets each including all n_samples.
  
  Arguments
  ---------
    
    n_samples: int
      The number of samples
  
  
  Returns
  -------
    
    cv: [(range(n_samples), range(n_samples))]
  
  """
    return [(range(n_samples), range(n_samples))]


def cvCustom(filename=None):
    """
  Read CV generator or train / test sets from a pickle file.
  """
    if filename == None:
        raise Exception("Error using cvCustom: filename is None")

    with open(filename, 'rb') as f:
        return pickle.load(f)


def cross_val_score(estimator,
                    X,
                    individual,
                    y=None,
                    scoring=None,
                    cv=None,
                    penalty=0.0,
                    fit_params=None):
    """
  Evaluate CV score for a particular individual.
  
  Arguments
  ---------
    
    estimator: estimator object implementing 'fit' and 'predict'
      The object to use to fit the data.
    
    X: array-like of shape (n_samples, n_features)
      The training input samples (correlations).
    
    individual: List[bool] of length n_features
      This is a boolean list of shape [n_features], in which an element is True 
      iff its corresponding feature is selected for retention.
    
    y: array-like of shape: (n_samples, 1)
      The target values (property values).
    
    scoring: string, callable or None, optional, default=None
      A string or a scorer callable object / function with signature 
      scorer(estimator, X, y). The parameter for sklearn.model_valdiation.cross_val_score,
      default = None, uses estimator.score().
    
    cv: cross-validation generator or an iterable
      Provides train/test splits
      
    penalty: float, optional, default=0.0
      The CV score is increased by 'penalty*sum(individual)'.
    
    fit_params: dict, optional,
      Parameters to pass to the fit method of the estimator.
  
  
  Returns
  -------
    
    score: float
      The cross validation score
  
  
  Notes
  -----
  
  Equivalent to:
  
    scores = cross_val_score(
      estimator,
      X[:,indices(individual)],
      y=y,
      scoring=scoring,
      cv=cv,
      fit_params=fit_params)
    return sqrt(np.mean(scores)) + penalty*sum(individual),
  """
    scores = sklearn.model_selection.cross_val_score(estimator,
                                                     X[:,
                                                       indices(individual)],
                                                     y=y,
                                                     scoring=scoring,
                                                     cv=cv,
                                                     fit_params=fit_params)
    return sqrt(np.mean(scores)) + penalty * sum(individual),



"""Plot casm.project.Prim structures"""
import numpy as np

def conventional_fcc_unit_cell():
    """Integer transformation matrix from primitive to conventional FCC cell

    Assumes primitive FCC lattice column vector matrix:

        [[   0, a/2,  a/2 ],
         [ a/2,   0,  a/2 ],
         [ a/2, a/2,    0 ]]

    Returns
    -------
    T: 3x3 np.array
        np.array([[ -1,  1,  1],
                  [  1, -1,  1],
                  [  1,  1, -1]])
    """
    return np.array([[-1, 1, 1], [1, -1, 1], [1, 1, -1]])

def conventional_bcc_unit_cell():
    """Integer transformation matrix from primitive to conventional BCC cell

    Assumes primitive BCC lattice column vector matrix:

        [[ -a/2,  a/2,  a/2 ],
         [  a/2, -a/2,  a/2 ],
         [  a/2,  a/2, -a/2 ]]

    Returns
    -------
    T: 3x3 np.array
        np.array([[ 0,  1,  1 ],
                  [ 1,  0,  1 ],
                  [ 1,  1,  0 ]])
    """
    return np.array([[0, 1, 1], [1, 0, 1], [1, 1, 0]])

def plot_prim(prim,
              super_cell=None,
              unit_cell=None,):
    if unit_cell is None:
        unit_cell = np.identity(3)
    if super_cell is None:
        super_cell = np.identity(3)
    T = np.dot(unit_cell, super_cell)
    a = 1.0
    L = np.array([[   0, a/2,  a/2 ],
                  [ a/2,   0,  a/2 ],
                  [ a/2, a/2,    0 ]])
    print(np.dot(L, T))
    return T.tolist()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import os
import sys
from contextlib import contextmanager
from io import StringIO
import six
if six.PY2:
    from mock import patch
else:
    from unittest.mock import patch


@contextmanager
def working_dir(wd):
    """
    Usage:
        # do some things in a different working directory, then return to the
        # the current working directory
        with working_dir(wd=some_path):
            ... some code ...
    
    Args:
        wd (str): working directory to use while in context
    """
    orig_wd = os.getcwd()
    if wd is None:
        wd = orig_wd
    os.chdir(wd)
    try:
        yield
    finally:
        os.chdir(orig_wd)


@contextmanager
def captured_output(wd=None):
    """
    Usage:
        # do some things and capture standard output and error
        # as StringIO objects sout, serr
        # optionally, pass 'wd', a different working directory in 
        # which to use while executing the code
        with captured_output(wd=some_path) as (sout, serr):
            ... some code ...
        # print sout and serr nicely:
        print_stringIO(sout)
        print_stringIO(serr)
    
    Args:
        wd (str): working directory to use while in context
    """
    with working_dir(wd):
        new_out, new_err = StringIO(), StringIO()
        old_out, old_err = sys.stdout, sys.stderr
        try:
            sys.stdout, sys.stderr = new_out, new_err
            yield sys.stdout, sys.stderr
        finally:
            sys.stdout, sys.stderr = old_out, old_err


def print_stringIO(strio):
    print("\n----\n", strio.getvalue(), "\n----")



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import ctypes
import glob
import json
import os
import six
from distutils.spawn import find_executable
from os.path import basename, dirname, join
from sys import platform

import sh


class API(object):
    """
    Class to provide access to the libccasm C API.

    Acts like a singleton by loading libcasm and libccasm into a class member the
    first time (and only the first time) that a new API instance is constructed.

    Each API instance uses the same shared class member to make calls.
    """
    class __API(object):
        """
        Hidden class of which there will be only one instance

        Attributes
        ----------

          lib_casm: ctypes.CDLL
            Handle to libcasm.*

          lib_ccasm: ctypes.CDLL
            Handle to libccasm.*

        """
        def __init__(self):
            """Loads dynamic libraries"""

            try:
                if 'LIBCASM' in os.environ:
                    libcasm_path = os.environ['LIBCASM']
                else:
                    casm_path = find_executable('ccasm')
                    if platform == 'darwin':
                        libcasm_path = sh.grep(sh.otool('-L', casm_path),
                                               'libcasm').split()[0]
                        libcasm_path = libcasm_path.replace(
                            '@loader_path', dirname(casm_path))
                    else:
                        libcasm_path = sh.grep(sh.ldd(casm_path),
                                               'libcasm').split()[2]
                libccasm_path = libcasm_path.replace('libcasm', 'libccasm')

                self.lib_casm = ctypes.CDLL(libcasm_path,
                                            mode=ctypes.RTLD_GLOBAL)
                self.lib_ccasm = ctypes.CDLL(libccasm_path,
                                             mode=ctypes.RTLD_GLOBAL)
            except Exception as e:
                print("\n~~~ Error loading casm libraries ~~~")
                if 'LIBCASM' in os.environ:
                    libcasm_path = os.environ['LIBCASM']
                    print("Looking for libcasm at LIBCASM:", libcasm_path)
                    if not os.path.exists(libcasm_path):
                        print("File does not exist")
                        print(
                            "Install CASM if it is not installed, or update your PATH, or set LIBCASM to the location of libcasm."
                        )
                    else:
                        print(
                            "File exists, but for unknown reason could not be loaded."
                        )
                else:
                    casm_path = find_executable('ccasm')
                    print("find_executable('ccasm'):", casm_path)
                    if casm_path is None:
                        print(
                            "Could not find 'ccasm' executable. CASM is not installed on your PATH."
                        )
                        print(
                            "Install CASM if it is not installed, or update your PATH, or set LIBCASM to the location of libcasm."
                        )
                    elif basename(dirname(casm_path)) == ".libs":
                        print(
                            "Found 'ccasm' executable in a '.libs' directory. Are you running tests?"
                        )
                        if platform == 'darwin':
                            check = join(dirname(casm_path), "libcasm.dylib")
                        else:
                            check = join(dirname(casm_path), "libcasm.so")
                        if os.path.exists(check):
                            print("You probably need to set LIBCASM=" + check)
                        else:
                            print(
                                "You probably need to re-make or update your PATH"
                            )
                    else:
                        print(
                            "Found 'ccasm', but for unknown reason could not determine libcasm location."
                        )
                        if platform == 'darwin':
                            print("otool -L:")
                            res = sh.otool('-L', casm_path)
                            for val in res:
                                print(val.strip())
                        else:
                            print("ldd:")
                            res = sh.ldd(casm_path)
                            for val in res:
                                print(val.strip())
                print("")
                raise e

            #### Argument types

            self.lib_ccasm.casm_STDOUT.restype = ctypes.c_void_p

            self.lib_ccasm.casm_STDERR.restype = ctypes.c_void_p

            self.lib_ccasm.casm_nullstream.restype = ctypes.c_void_p

            self.lib_ccasm.casm_ostringstream_new.restype = ctypes.c_void_p

            self.lib_ccasm.casm_ostringstream_delete.argtypes = [
                ctypes.c_void_p
            ]
            self.lib_ccasm.casm_ostringstream_delete.restype = None

            self.lib_ccasm.casm_ostringstream_size.argtypes = [ctypes.c_void_p]
            self.lib_ccasm.casm_ostringstream_size.restype = ctypes.c_ulong

            self.lib_ccasm.casm_ostringstream_strcpy.argtypes = [
                ctypes.c_void_p,
                ctypes.POINTER(ctypes.c_char)
            ]
            self.lib_ccasm.casm_ostringstream_strcpy.restype = ctypes.POINTER(
                ctypes.c_char)

            self.lib_ccasm.casm_primclex_null.argtypes = None
            self.lib_ccasm.casm_primclex_null.restype = ctypes.c_void_p

            self.lib_ccasm.casm_primclex_new.argtypes = [
                ctypes.c_char_p, ctypes.c_void_p, ctypes.c_void_p
            ]
            self.lib_ccasm.casm_primclex_new.restype = ctypes.c_void_p

            self.lib_ccasm.casm_primclex_delete.argtypes = [ctypes.c_void_p]
            self.lib_ccasm.casm_primclex_delete.restype = None

            self.lib_ccasm.casm_primclex_refresh.argtypes = [
                ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p,
                ctypes.c_bool, ctypes.c_bool, ctypes.c_bool, ctypes.c_bool,
                ctypes.c_bool
            ]
            self.lib_ccasm.casm_primclex_refresh.restype = None

            self.lib_ccasm.casm_command_list.argtypes = [ctypes.c_void_p]
            self.lib_ccasm.casm_command_list.restype = None

            self.lib_ccasm.casm_capi.argtypes = [
                ctypes.c_char_p, ctypes.c_void_p, ctypes.c_char_p,
                ctypes.c_void_p, ctypes.c_void_p
            ]
            self.lib_ccasm.casm_capi.restype = ctypes.c_int

    __api = None

    def __init__(self):
        """
        Acts like a singleton by loading libcasm and libccasm into a class member the
        first time (and only the first time) that a new API instance is constructed.

        Each API instance uses the same shared class member to make calls.
        """
        if API.__api is None:
            API.__api = API.__API()
        return

    def stdout(self):
        """
        Get a pointer to a CASM::Log that writes to std::cout.

        This does not need to be deleted manually.
        """
        return API.__api.lib_ccasm.casm_STDOUT()

    def stderr(self):
        """
        Get a pointer to a CASM::Log that writes to std::cerr

        This does not need to be deleted manually.
        """
        return API.__api.lib_ccasm.casm_STDERR()

    def nullstream(self):
        """
        Get a pointer to a CASM::Log that writes to null.

        This does not need to be deleted manually.
        """
        return API.__api.lib_ccasm.casm_nullstream()

    def ostringstream_new(self):
        """
        Construct a CASM::OStringStreamLog that writes to std::ostringstream.

        Returns
        -------
          ptr: CASM::OStringStreamLog pointer
            Used for capturing CASM output in a string.
            This ptr needs to be deleted manually by using API.ostringstream_delete(ptr)
        """
        return API.__api.lib_ccasm.casm_ostringstream_new()

    def ostringstream_to_str(self, ptr):
        """
        Copy the value of a CASM::OStringStreamLog to a Python string.

        Arguments
        ---------

          ptr: CASM::OStringStreamLog pointer


        Returns
        -------
          s: str
            The contents of a CASM::OStringStreamLog

        """
        c_str = ctypes.create_string_buffer(
            API.__api.lib_ccasm.casm_ostringstream_size(ptr))
        API.__api.lib_ccasm.casm_ostringstream_strcpy(ptr, c_str)
        return c_str.value

    def ostringstream_delete(self, ptr):
        """
        Delete a CASM::OStringStreamLog

        Arguments
        ---------

          ptr: CASM::OStringStreamLog pointer

        """
        API.__api.lib_ccasm.casm_ostringstream_delete(ptr)
        return

    def primclex_null(self):
        """
        Construct a CASM::PrimClex nullptr

        Returns
        -------
          ptr: CASM::PrimClex nullptr

        """
        return API.__api.lib_ccasm.casm_primclex_null()

    def primclex_new(self, path, log, err_log):
        """
        Construct a new CASM::PrimClex

        Arguments
        ---------

          path: str
            The path to a CASM project

          log: CASM::Log pointer
            A pointer to a CASM::Log to write standard output

          err_log: CASM::Log pointer
            A pointer to a CASM::Log to write error output


        Notes
        -------

        CASM::Log pointers can be obtained from:
          1) API.stdout()
          2) API.stderr()
          3) API.nullstream()
          4) API.ostringstream()


        Returns
        -------
          ptr: CASM::PrimClex pointer

        """
        return API.__api.lib_ccasm.casm_primclex_new(six.b(path), log, err_log)

    def primclex_refresh(self,
                         ptr,
                         log,
                         err_log,
                         read_settings=False,
                         read_composition=False,
                         read_chem_ref=False,
                         read_configs=False,
                         clear_clex=False):
        """
        Call CASM::PrimClex::refresh to reload PrimClex data from settings files

        Arguments
        ---------

          ptr: CASM::PrimClex pointer

          log: CASM::Log pointer
            A pointer to a CASM::Log to write standard output

          err_log: CASM::Log pointer
            A pointer to a CASM::Log to write error output

          read_settings: bool, optional, default=False
            If True, read project_settings.json

          read_composition: bool, optional, default=False
            If True, read composition_axes.json

          read_chem_ref: bool, optional, default=False
            If True, read chemical_reference.json

          read_configs: bool, optional, default=False
            If True, read SCEL and config_list.json

          clear_clex: bool, optional, default=False
            If True, clear stored orbitrees, clexulators, and eci


        Note
        ----

          This does not check if what you request will cause problems.

        """
        API.__api.lib_ccasm.casm_primclex_refresh(ptr, log, err_log,
                                                  read_settings,
                                                  read_composition,
                                                  read_chem_ref, read_configs,
                                                  clear_clex)
        return

    def primclex_delete(self, ptr):
        """
        Delete a CASM::PrimClex

        Arguments
        ---------

          ptr: CASM::PrimClex pointer

        """
        API.__api.lib_ccasm.casm_primclex_delete(ptr)
        return

    def command_list(self):
        """
        Get list of recognized casm commands implemented at the libcasm level

        Returns
        -------
          s: str
            JSON array containing the list of recognized casm commands

        """
        ptr = self.ostringstream_new()
        API.__api.lib_ccasm.casm_command_list(ptr)
        s = self.ostringstream_to_str(ptr)
        self.ostringstream_delete(ptr)
        return s

    def capi(self, args, primclex, root, log, err_log):
        """
        Make an API call

        Arguments
        ---------

          args: str
            A string containing the arguments for the casm command to be executed.

              Ex: "select --set-on -o /abspath/to/my_selection"
              Ex: "query -k 'configname selected' -v -o STDOUT"

          primclex: CASM::PrimClex pointer
            A pointer to a CASM::PrimClex, as obtained from API.primclex_new()

          root: str
            A string giving the path to a root directory of a CASM project, typically
            casm.project.Project.path

          log: CASM::Log pointer
            A pointer to a CASM::Log to write standard output

          err_log: CASM::Log pointer
            A pointer to a CASM::Log to write error output

        Returns
        -------
          returncode: int

            Possible values:

              0: No error

              1: ERR_INVALID_ARG
                Command line input is not valid

              2: ERR_UNKNOWN
                Misc. or Unknown error

              3: ERR_NO_PROJ
                No CASM project can be found in expected location

              4: ERR_INVALID_INPUT_FILE
                An expected input file is invalid

              5: ERR_MISSING_INPUT_FILE
                An expected input file can not be found

              6: ERR_EXISTING_FILE
                A file might be overwritten

              7: ERR_MISSING_DEPENDS
                Requested command can not be performed because some dependency needs
                to be done first (i.e. no basis set, so can't use clexulator)

              8: ERR_OTHER_PROJ
                Unknown attempting to overwrite another CASM project

        """
        return API.__api.lib_ccasm.casm_capi(six.b(args), primclex,
                                             six.b(root), log, err_log)


def _project_path(dir=None):
    """
    Crawl up from dir to find '.casm'. If found returns the directory containing the '.casm' directory.
    If not found, return None.

    Args:
    If dir == None, set to os.getcwd()
    """
    if dir == None:
        dir = os.getcwd()
    else:
        dir = os.path.abspath(dir)
    if not os.path.isdir(dir):
        raise Exception("Error, no directory named: " + dir)
    curr = dir
    cont = True
    while cont == True:
        test_path = os.path.join(curr, ".casm")
        if os.path.isdir(test_path):
            return curr
        elif curr == os.path.dirname(curr):
            return None
        else:
            curr = os.path.dirname(curr)
    return None


def command_list():
    """
    Get list of recognized casm commands implemented at the libcasm level

    Returns
    -------
      s: str
        JSON array containing the list of recognized casm commands

    """
    _api = API()
    return _api.command_list()


def casm_command(args, root=None, combine_output=False):
    """
    Execute a command via the c api, writing output to stdout/stderr. If required by the specific
    command, a temporary PrimClex instance will be constrcuted. Use Project.command or
    Project.capture to avoid re-initializations.

    Arguments
    ---------

      args: str
        A string containing the arguments for the casm command to be executed.

          Ex: "select --set-on -o /abspath/to/my_selection"
          Ex: "query -k 'configname selected' -v -o STDOUT"

      root: str (optional)
        A string giving the path to a root directory of a CASM project,
        typically casm.project.Project.path. If None given, a `'.casm'`
        directory is searched for starting at the current working directory,
        and if not found the empty string is used.

      combine_output: bool (optional, default=False)
        If True, print stdout and stderr to same str and only ret

    Returns
    -------
      returncode: The result of running the command via the command line iterface. 'stdout' and
        'stderr' are in text type ('unicode'/'str').
    """
    _api = API()

    # set default root
    if root is None:
        root = _project_path()
        if root is None:
            root = ""

    # construct stringstream objects to capture stdout, stderr
    ss = _api.stdout()
    if combine_output:
        ss_err = ss
    else:
        ss_err = _api.stderr()

    returncode = _api.capi(args, _api.primclex_null(), root, ss, ss_err)

    return returncode


def casm_capture(args, root=None, combine_output=False):
    """
    Execute a command via the c api and store stdout/stderr result as str. If required by the
    specific command, a temporary PrimClex instance will be constrcuted. Use Project.command or
    Project.capture to avoid re-initializations.

    Arguments
    ---------

      args: str
        A string containing the arguments for the casm command to be executed.

          Ex: "select --set-on -o /abspath/to/my_selection"
          Ex: "query -k 'configname selected' -v -o STDOUT"

      root: str (optional)
        A string giving the path to a root directory of a CASM project,
        typically casm.project.Project.path. If None given, a `'.casm'`
        directory is searched for starting at the current working directory,
        and if not found the empty string is used.

      combine_output: bool (optional, default=False)
        If True, print stdout and stderr to same str and only ret

    Notes
    -------
      Prefer not to use this function if you have an existing casm.project.Project instance that
      the command will operate on. Instead use Project.command() which will execute the command and
      then update the Project instance's ProjectSettings and DirectoryStructure to reflect any
      changes that have occurred.

    Returns
    -------
      (stdout, stderr, returncode): The result of running the command via the
          command line iterface. 'stdout' and 'stderr' are in text type ('unicode'/'str'). If
          'combine_output' is True, then returns (combined_output, returncode).

    """
    _api = API()

    # set default root
    if root is None:
        root = _project_path()
        if root is None:
            root = ""

    # construct stringstream objects to capture stdout, stderr
    ss = _api.ostringstream_new()
    if combine_output:
        ss_err = ss
    else:
        ss_err = _api.ostringstream_new()

    returncode = _api.capi(args, _api.primclex_null(), root, ss, ss_err)

    # copy strings and delete stringstreams
    stdout = _api.ostringstream_to_str(ss)
    _api.ostringstream_delete(ss)

    if combine_output:
        res = (stdout.decode('utf-8'), returncode)
    else:
        stderr = _api.ostringstream_to_str(ss_err)
        _api.ostringstream_delete(ss_err)

        res = (stdout.decode('utf-8'), stderr.decode('utf-8'), returncode)

    return res



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import json
import sys

from casm.misc import compat, noindent
import casm.vaspwrapper


def main():
    print("Begin vasp.relax.report")

    if len(sys.argv) != 2:
        print("Usage: vasp.relax.report configdir")
        sys.exit()

    configdir = sys.argv[1]

    try:
        output = casm.vaspwrapper.Relax.properties(configdir)
    except:
        print((
            "Unable to report properties for directory %s. Please verify that it contains a completed VASP calculation."
            % configdir))
        raise

    compat.dump(json,
                output,
                'properties.calc.json',
                'w',
                cls=noindent.NoIndentEncoder,
                indent=4,
                sort_keys=True)

    print("Finish vasp.relax.report\n\n")


if __name__ == "__main__":
    main()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

# conda's current version of pandas raises these warnings, but they are safe
# see: https://stackoverflow.com/questions/40845304
import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

import copy
from io import StringIO
import json
import os
import subprocess

import numpy as np
import pandas
import six

from casm.project.project import Project
from casm.project.query import query
from casm.misc import compat


class Selection(object):
    """
    A Selection object contains information about a CASM project

    Attributes
    ----------

      proj: casm.Project, optional, default=Project containing the current working directory
        the CASM project the selection belongs to

      path: string, optional, default="MASTER"
        path to selection file, or "MASTER" (Default="MASTER")

      type: string, optional, default="config"
        type of selected object: "config" or "scel"

      all: bool, optional, default=True
        if True, self.data will include all configurations, whether selected or
        not. If False, only selected configurations will be included.

      data: pandas.DataFrame
        A pandas.DataFrame describing the selected configurations. Has at least
        'configname' and 'selected' (automatically converted to bool) columns.

    """
    def __init__(self, proj=None, path="MASTER", type="config", all=True):
        """
        Construct a CASM Project representation.

        Arguments
        ---------

          proj: casm.Project, optional, default=Project containing the current working directory
            the CASM project the selection belongs to

          path: string, optional, default="MASTER"
            path to selection file, or "MASTER" (Default="MASTER")

          type: string, optional, default="config"
            type of selected object: "config" or "scel"

          all: bool, optional, default=True
            if True, self.data will include all configurations, whether selected or
            not. If False, only selected configurations will be included.


        """
        if proj == None:
            proj = Project()
        elif not isinstance(proj, Project):
            raise Exception(
                "Error constructing Selection: proj argument is not a CASM project"
            )
        self.proj = proj

        self.path = path
        if os.path.isfile(path):
            self.path = os.path.abspath(path)

        self.type = type
        self.all = all

        self._data = None

        # reserved for use by casm.plotting
        self.src = None

    @property
    def data(self):
        """
        Get Selection data as a pandas.DataFrame

        If the data is modified, 'save' must be called for CASM to use the modified selection.
        """
        if self._data is None:
            if self.path in ["MASTER", "ALL", "CALCULATED"]:
                self._data = query(self.proj, ['name', 'selected'],
                                   self.path,
                                   self.type,
                                   verbatim=True,
                                   all=self.all)
            elif self._is_json():
                self._data = pandas.read_json(self.path, orient='records')
            else:
                with open(self.path, compat.pandas_rmode()) as f:
                    if compat.peek(f) == '#':
                        f.read(1)
                    self._data = pandas.read_csv(f,
                                                 sep=compat.str(' +'),
                                                 engine='python')

            self._clean_data()

            if not self.all:
                self._data = self._data[self._data['selected'] == True]

        return self._data

    @data.setter
    def data(self, input):
        self._data = input

    def save(self, data=None, force=False):  #make it generalized ##todo
        """Save the current selection.

        This method also allows completely replacing the 'data' describing the selected configurations.

        Arguments
        ---------
        data: pandas.DataFrame
            If provided, it replaces `self.data` before writing. It must have `"name"` and `"selected"` columns to be opened later.

        force: bool
            Force overwrite existing files

        Raises
        ------
        Exception
            - If attempting to overwrite an existing file and `force==False`
            - If `self.path in ["ALL", "CALCULATED"]`
        """
        if self.path == "MASTER":

            self.path = self.proj.dir.master_selection(self.type)
            self.save(data=data, force=force)
            self.path = "MASTER"

        elif self.path in ["ALL", "CALCULATED"]:
            raise Exception("Cannot save the '" + self.path + "' Selection")

        else:

            if data is not None:
                self._data = data.copy()
                self._clean_data()

            if self._data.columns[0] != "name":
                raise Exception(
                    "The first column in Selection.data must be 'name'")
            if self._data.columns[1] != "selected":
                raise Exception(
                    "The second column in Selection.data must be 'selected'")

            if os.path.exists(self.path) and not force:
                raise Exception("File: " + self.path + " already exists")

            backup = self.path + ".tmp"
            if os.path.exists(backup):
                raise Exception("File: " + backup + " already exists")

            if self._is_json():
                self._data.to_json(backup, orient='records')
            else:
                self.astype("selected", np.int_)
                with open(backup, compat.pandas_wmode()) as f:
                    f.write('# ')  # TODO: make this optional
                    self._data.to_csv(f, sep=compat.str(' '), index=False)
                self.astype("selected", bool)
            os.rename(backup, self.path)

    def saveas(self, path="MASTER", force=False):
        """
        Create a new Selection from this one, save and return it

        Arguments
        ---------
        path: str
            Path to a selection file or standard selection name.

        force: bool
            Force overwrite existing files

        Returns
        -------
        sel: Selection
            The new Selection created from this one
        """
        sel = Selection(self.proj, path, type=self.type, all=self.all)
        sel._data = self.data.copy()
        sel.save(force=force)
        return sel

    def _is_json(self):
        return self.path[-5:].lower() == ".json"

    def astype(self, columnname, dtype):
        """Convert a column to another type

        Example: Convert column of int (0 / 1) to bool:

            selection.astype('selected', bool)

        Arguments
        ---------
        columns: List(str)
            Data requested, will be added as columns in `self.data`. This corresponds to the `-k` option of `casm query`. A list of options can be obtained from `casm query --help properties`.

        force: bool
            If `force==False`, input `columns` that already exist in `self.data.columns` will be ignored and those columns will not be updated. If `force==True`, those columns will be overwritten with new data.

        """
        self._data.loc[:,
                       columnname] = self._data.loc[:,
                                                    columnname].astype(dtype)

    def _clean_data(self):
        self.astype('selected', bool)

    def query(self, columns, force=False, verbose=False):
        """ Query requested columns and store them in 'data'.

        Will not overwrite columns that already exist, unless 'force'==True. Will query data for all configurations, whether selected or not, if `self.all == True`.

        Arguments
        ---------
        columns: List(str)
            Data requested, will be added as columns in `self.data`. This corresponds to the `-k` option of `casm query`. A list of options can be obtained from `casm query --help properties`.

        force: bool
            If `force==False`, input `columns` that already exist in `self.data.columns` will be ignored and those columns will not be updated. If `force==True`, those columns will be overwritten with new data.

        verbose: bool
            How much to print to stdout.
        """

        if force == False:
            _col = [x for x in columns if x not in self.data.columns]
        else:
            _col = columns

        if verbose:
            print("# Query requested:", columns)
            if force == False:
                print("# Use existing:",
                      [x for x in columns if x in self.data.columns])
            else:
                print("# Overwrite existing:",
                      [x for x in columns if x in self.data.columns])
            if len(_col) == 0:
                print("# No query necessary")
            else:
                print("# Querying:", _col)

        if len(_col) == 0:
            return

        df = query(self.proj,
                   _col,
                   self.path,
                   self.type,
                   verbatim=True,
                   all=self.all)

        if verbose:
            print("#   DONE\n")

        msg = "querying different numbers of records: {0}, {1}".format(
            self.data.shape, df.shape)
        assert self.data.shape[0] == df.shape[0], msg

        for c in df.columns:
            self.data.loc[:, c] = df.loc[:, c].values

    def write_pos(self, all=False):
        """Write POS file for selected configurations

        Equivalent to `casm query -c <selection> --write-pos`. This writes a POS file for selected configurations at `<root>/training_data/<configname>/POS`.

        Arguments
        ---------

        all: bool, optional, default=False
            If True, will write POS file for all configurations in the selection whether selected or not. If False, only write POS file for selected configurations.
        """
        self.proj.capture("query -c " + self.path + " --write-pos")

    def add_data(self, name, data=None, force=False):
        """Add selection data, either by query or from an existing DataFrame

        Equivalent to:

            if name not in sel.data.columns or force == True:
                if data is None:
                    sel.query([name], force)
                else:
                    sel.data.loc[:,name] = data

        Arguments
        ---------
        columns: List(str)
            The data requested. This corresponds to the `-k` option of `casm query`, or a column in `data`. A list of options can be obtained from `casm query --help properties`.

        data: pandas.DataFrame
            May be provided as a data source as an alternative to `casm query`.

        force: bool
            If True, overwrite existing data in `self.data`.
        """
        if name not in self.data.columns or force == True:
            if data is None:
                self.query([name], force)
            else:
                self.data.loc[:, name] = data



class DofClassError(Exception):
    """Exception handling"""
    def __init__(self, message):
        self.message = message

    def __str__(self):
        """ Writes out the error message

        Returns
        -------
        string

        """
        return self.message


class CmagspinAttr:
    """Class containing information specific to Cmagspin dof.
       This object will be constructed from casm.project.attribute_info.AttributeInfo class
       which digests its information.

       self.atom_props: List[Dict] - Contains the list of atom properites (with atom name and it's value)
       Look at the example below for detailed description of object properties

       Consider the example of NaFeO2 with Fe having a +5 magnetic moment and rest all being 0
       self.atom_props: [{"site_index":0, "atom": "Na", "value":0},{"site_index":1,"atom":"Fe", "value":5},{"site_index":2, "atom":"O","value":0},{"site_index":3, "atom":"O","value":0}]"""
    def __init__(self, dof_info):
        """Constructs the CmagspinAttr object from AttributeInfo object

        Parameters
        ----------
        dof_info : casm.project.attribute_info.AttributeInfo

        """
        try:
            self.atom_props = [{
                "site_index":
                x,
                "atom":
                dof_info.atom_type[x],
                "value":
                dof_info.atom_dofs["Cmagspin"]["value"][x]
            } for x in range(0, len(dof_info.atom_type))]
        except:
            raise DofClassError(
                "Could not construct CmagspinAttr class!! Check if you're dealing with Cmagspin dof calculations"
            )

    def vasp_input_tags(self, sort=True):
        """Returns a dictionary of VASP input tags specific to collinear magnetic spin calculations.
        The collinear magnetic spin specific tags are as follows:
            
        MAGMOM, ISPIN

        Parameters
        ----------
        sort: bool (This should match the sort used to write POSCAR file (whether the basis atoms are sorted))

        Returns
        -------
        dict{"vasp_input_tag": "value"}

        """
        #TODO: Group together atoms of same MAGMOM together
        #TODO: Also add ISPIN default tag which is required if missed in INCAR.base

        if sort is True:
            self.atom_props.sort(key=lambda x: x["atom"])

        magmom_value = ""
        for atom_props in self.atom_props:
            magmom_value = magmom_value + str(atom_props["value"][0]) + " "

        tags = dict()
        tags["MAGMOM"] = magmom_value
        return tags

    def vasp_output_dictionary(self, outcar, sort=True):
        """Returns the attribute specific vasp output dictionary which can be updated
        to the whole output dictionary which will be printed as properties.calc.json.

        For Cmagspin, this will be magnetic moment of each individual species

        Parameters
        ----------
        outcar : casm.vasp.io.outcar (Class containing information about magmom of individual species from OUTCAR)
        sort : bool (This should be the same sort used while writing POSCAR)

        Returns
        -------
        dict{"Cmagspin":{"value":[list]}}

        """
        if sort is True:
            self.atom_props.sort(key=lambda x: x["atom"])

        permutation_vector_to_unsort = [
            x["site_index"] for x in self.atom_props
        ]

        output = {}
        output["Cmagspin"] = {}
        output["Cmagspin"]["value"] = [[mag] for site_index, mag in sorted(
            zip(permutation_vector_to_unsort, outcar.mag))]

        return output



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

# import warnings
# warnings.filterwarnings("ignore", category=DeprecationWarning)
# warnings.filterwarnings("ignore", message="numpy.dtype size changed")
# warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

import argparse
import json
import os
import pty
import subprocess
import sys

import casm.scripts.casm_calc
import casm.scripts.casm_learn
from casm.api import command_list, casm_command


def _exec(argv=None):
    subprocess.Popen(argv)


def _shell(argv=None):
    orig = os.environ['PS1']
    os.environ['PS1'] = "<shell> " + os.environ['PS1']
    pty.spawn('/bin/bash')
    os.environ['PS1'] = orig


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    libcasm_commands = json.loads(command_list())
    if '--version' in libcasm_commands:
        libcasm_commands.remove('--version')
    found_libcasm = True

    python_commands = {
        'calc': casm.scripts.casm_calc.main,
        'learn': casm.scripts.casm_learn.main,
        'shell': _shell,
        'exec': _exec
    }

    commands = sorted(libcasm_commands + list(python_commands.keys()))

    # if not found_libcasm:
    #     print("Could not find libcasm. Please check your installation.")
    #     return 1

    parser = argparse.ArgumentParser(
        description='CASM: First-principles based statistical mechanics')
    parser.add_argument('command',
                        help="CASM command to execute",
                        type=str,
                        default="",
                        nargs="?",
                        metavar="<command>")
    parser.add_argument('args',
                        help="CASM command arguments",
                        nargs=argparse.REMAINDER,
                        metavar="...")
    parser.add_argument('--desc',
                        help="Print command list",
                        action="store_true",
                        default=False)
    parser.add_argument('--version',
                        help="Print casm version",
                        action="store_true",
                        default=False)
    parser.add_argument(
        '--path',
        help=
        "Path to project. Default uses project containing current working directory.",
        type=str,
        default="")

    args = parser.parse_args(argv)

    if args.version:
        args.command = "version"
        args.args = []
    if args.desc:
        parser.print_help()
        print("")
        print("available commands:")
        for cmd in commands:
            print("  " + cmd)
        print("")
        print("For help using a command: 'casm <command> --help'\n")
        print("For step by step help use: 'casm status -n'\n")
    elif args.command in python_commands:
        python_commands[args.command](argv[1:])
    elif args.command in libcasm_commands:
        args.args = ["\'" + x + "\'" for x in args.args]
        res = casm_command(' '.join([args.command] + args.args),
                           root=args.path)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



""" Job manipuation routines for VASP"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import os
import shutil
import re
import subprocess
import sys
import time
import gzip
import warnings
import signal

from casm.vasp.error import VaspError, VaspWarning, error_check, crash_check
from casm.vasp import io


def complete_job(jobdir, settings):
    """Remove files from a vasp job directory

       Args:
         jobdir: path to current job directory
       Settings:
         copy: Does nothing
         move: Does nothing
         compress: Compresses listed files
         backup: Does nothing
         remove: Deletes listed files
    """

    print("Complete VASP job: " + jobdir)
    sys.stdout.flush()

    # remove files
    print("  rm:", end=' ')
    for f in settings["remove"]:
        if not f in (settings["copy"] + settings["move"] +
                     settings["compress"] + settings["backup"]):
            if os.path.isfile(os.path.join(jobdir, f)):
                print(f, end=' ')
                os.remove(os.path.join(jobdir, f))
    for f in settings["extra_input_files"]:
        if os.path.isfile(os.path.join(jobdir, f)):
            print(f, end=' ')
            os.remove(os.path.join(jobdir, f))
    print("")

    # compress files
    print(" gzip:", end=' ')
    for file in settings["compress"]:
        if os.path.isfile(os.path.join(jobdir, file)):
            print(file, end=' ')
            # Open target file, target file.gz
            f_in = open(os.path.join(jobdir, file), 'rb')
            f_out = gzip.open(os.path.join(jobdir, file) + '.gz', 'wb')
            # Compress, close files
            f_out.writelines(f_in)
            f_out.close()
            f_in.close()
            # Remove original target file
            os.remove(os.path.join(jobdir, file))
    print("")
    print("")
    sys.stdout.flush()


def run(jobdir=None,
        stdout="std.out",
        stderr="std.err",
        npar=None,
        ncore=None,
        command=None,
        ncpus=None,
        kpar=None,
        poll_check_time=5.0,
        err_check_time=60.0,
        err_types=None):
    """ Run vasp using subprocess.

        The 'command' is executed in the directory 'jobdir'.

        Args:
            jobdir:     directory to run vasp.  If jobdir is None, the current directory is used.
            stdout:     filename to write to.  If stdout is None, "std.out" is used.
            stderr:     filename to write to.  If stderr is None, "std.err" is used.
            npar:       (int or None) VASP INCAR NPAR setting. If npar is None, then NPAR is removed from INCAR
            kpar:       (int or None) VASP INCAR KPAR setting. If kpar is None, then KPAR is removed from INCAR
            ncore:      (int or None) VASP INCAR NCORE setting. If not npar is None or ncore is None, then NCORE is removed from INCAR
            command:    (str or None) vasp execution command
                        If command != None: then 'command' is run in a subprocess
                        Else, if ncpus == 1, then command = "vasp"
                        Else, command = "mpirun -np {NCPUS} vasp"
            ncpus:      (int) if '{NCPUS}' is in 'command' string, then 'ncpus' is substituted in the command.
                        if ncpus==None, $PBS_NP is used if it exists, else 1
            poll_check_time: how frequently to check if the vasp job is completed
            err_check_time: how frequently to parse vasp output to check for errors
            err_types:  List of error types to check for. Supported errors: 'IbzkptError', 'SubSpaceMatrixError', 'NbandsError'. Default: None, in which case only SubSpaceMatrixErrors are checked.

    """
    print("Begin vasp run:")
    sys.stdout.flush()

    if jobdir is None:
        jobdir = os.getcwd()

    currdir = os.getcwd()
    os.chdir(jobdir)

    if ncpus is None:
        if "PBS_NP" in os.environ:
            ncpus = os.environ["PBS_NP"]
        elif "SLURM_NTASKS" in os.environ:
            ncpus = os.environ["SLURM_NTASKS"]
        else:
            ncpus = 1

    if command is None:
        if ncpus == 1:
            command = "vasp"
        else:
            command = "mpirun -np {NCPUS} vasp"

    if re.search("\{NCPUS\}", command):
        command = command.format(NCPUS=str(ncpus))

    ### Expand remaining environment variables
    command = os.path.expandvars(command)

    if npar is not None:
        ncore = None

    if npar is not None or ncore is not None:
        io.set_incar_tag({"NPAR": npar, "NCORE": ncore}, jobdir)

    if kpar is not None:
        io.set_incar_tag({"KPAR": kpar}, jobdir)

    print("  jobdir:", jobdir)
    print("  exec:", command)
    sys.stdout.flush()

    sout = open(os.path.join(jobdir, stdout), 'w')
    serr = open(os.path.join(jobdir, stderr), 'w')
    err = None
    p = subprocess.Popen(command.split(), stdout=sout, stderr=serr)

    # wait for process to end, and periodically check for errors
    poll = p.poll()
    last_check = time.time()
    stopcar_time = None
    while poll is None:
        time.sleep(poll_check_time)

        if time.time() - last_check > err_check_time:
            last_check = time.time()
            err = error_check(jobdir, os.path.join(jobdir, stdout), err_types)
            if err != None:
                # FreezeErrors are fatal and usually not helped with STOPCAR
                if "FreezeError" in err.keys():
                    print("  VASP is frozen, killing job")
                    sys.stdout.flush()
                    # Sometimes p.kill doesn't work if the process is on multiple nodes
                    os.kill(p.pid, signal.SIGKILL)
                    p.kill()
                    # If the job is re-invoked (e.g. via mpirun or srun) too quickly
                    #   after the previous job ended, infinitiband clusters can have
                    #   some issues with resource allocation. A 30s sleep solves this.
                    time.sleep(30)
                # Other errors can be killed with STOPCAR, which is safer
                elif stopcar_time is None:
                    print("  Found errors:", end=' ')
                    for e in err:
                        print(e, end=' ')
                    print("\n  Killing job with STOPCAR")
                    sys.stdout.flush()
                    io.write_stopcar('e', jobdir)
                    stopcar_time = time.time()
                    time.sleep(30)
                # If the STOPCAR exists, wait 5 min before manually killing the job
                elif time.time() - stopcar_time > 300:
                    print("  VASP is non-responsive, killing job")
                    sys.stdout.flush()
                    os.kill(p.pid, signal.SIGKILL)
                    p.kill()
                    # If the job is re-invoked (e.g. via mpirun or srun) too quickly
                    #   after the previous job ended, infinitiband clusters can have
                    #   some issues with resource allocation. A 30s sleep solves this.
                    time.sleep(30)

        poll = p.poll()

    # close output files
    sout.close()
    serr.close()

    os.chdir(currdir)

    print("Run complete")
    sys.stdout.flush()

    # check finished job for errors
    if err is None:
        # Crash-type errors take priority over any other error that may show up
        err = crash_check(jobdir, os.path.join(jobdir, stdout), err_types)
        if err is None:
            err = error_check(jobdir, os.path.join(jobdir, stdout), err_types)
    if err != None:
        print("  Found errors:", end=' ')
        for e in err:
            print(e, end=' ')
    print("\n")

    return err



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import sklearn.linear_model


class SelectionWrapper(object):
    """ A wrapper function to handle selectors like RandomizedLasso,
    which cannot parse an "estimator" input but otherwise have
    the appropriate properties

    *** ONLY USE THIS IF YOU KNOW WHAT YOU ARE DOING ***
    """
    def __new__(cls, estimator, name=None, **kwargs):
        if hasattr(sklearn.linear_model, name):
            return getattr(sklearn.linear_model, name)(**kwargs)
        raise AttributeError("Could not find: " + name)

    def __init__(self):
        pass



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import copy
from casm.learn import empty_individual
from casm.learn.model_selection import cross_val_score
from casm.learn.fit import open_halloffame, make_fitting_data, make_estimator, \
  add_individual_detail, print_halloffame


def make_population(n_features, input_options):
    """
  Read direct selection population options and use to construct a population of
  individuals to be fit.
  
  Arguments
  ---------
    
    n_features: int
      The size of the individuals to be constructed
    
    input_options: iterable of dict
      Each dict contains options specifying an individual or individuals to be
      included in the population. Three forms are recognized:
      
      bitstring:
        Ex.: {"bitstring" : "01110001100..."}
        The string may have fewer than n_features digits, but not more. If fewer,
        the rest are assumed to be '0'.
      
      indices:
        Ex.: {"indices" : [1, 2, 3, 7, 8]}
        Specify which features should be 'on' by index.
      
      from_halloffame:
        Ex.: {"from_halloffame" : "my_halloffame.pkl", "individuals" : [0, 2, 5]}
        Specifies a hall of fame and particular individuals (by index) in the hall
        of fame to include in the population. "individuals" is optional, with the
        default behaviour including all individuals in the hall of fame.
  
  Returns
  -------
    
    population: List[individuals]
      A list of the individuals to be fit
    
  """
    population = []
    for opt in input_options:

        if "bitstring" in opt:
            indiv = empty_individual(n_features)
            for index, bit in enumerate(opt["bitstring"]):
                if int(bit):
                    indiv[index] = True
            if sum(indiv) == 0:
                raise Exception(
                    "Error using making individual from bitstring: No basis functions selected"
                )
            population.append(indiv)

        elif "indices" in opt:
            indiv = empty_individual(n_features)
            for index in opt["indices"]:
                indiv[index] = True
            if sum(indiv) == 0:
                raise Exception(
                    "Error using making individual from bitstring: No basis functions selected"
                )
            population.append(indiv)

        elif "from_halloffame" in opt:
            hall = open_halloffame(opt["from_halloffame"])
            if "individuals" in opt:
                for index in opt["individuals"]:
                    population.append(hall[index])
            else:
                for indiv in hall:
                    population.append(indiv)

        else:
            print("Unrecognized option:\n")
            print(opt)
            raise Exception(
                "Error making DirectSelection population: unrecognized option")

    return population


def direct_fit(input, save=True, verbose=True, read_existing=True, hall=None):
    """
  Fit ECI and add details for a set of individuals specified for feature_selection
  method 'DirectSelection' via the 'population' kwarg.
  
  Arguments
  ---------
    
    input: dict
      The input settings as a dict
    
    save: boolean, optional, default=True
      Save a pickle file containing the training data and scoring metric. The file
      name, which can be specified by input["fit_data_filename"], defaults to "fit_data.pkl".
    
    verbose: boolean, optional, default=True
      Print information to stdout.
    
    read_existing: boolean, optional, default=True
      If it exists, read the pickle file containing the training data and scoring 
      metric. The file name, which can be specified by input["fit_data_filename"], 
      defaults to "fit_data.pkl".
    
    hall: deap.tools.HallOfFame, optional, default=None
      A Hall Of Fame to add resulting individuals to
    
  
  Returns
  -------
    
    population: List[individuals]
      A list of the individuals that were fit
    
  """
    # construct FittingData
    fdata = make_fitting_data(input,
                              save=True,
                              verbose=verbose,
                              read_existing=True)

    kwargs = input["feature_selection"]["kwargs"]

    population = make_population(fdata.n_features, kwargs["population"])

    for indiv_i, indiv in enumerate(population):

        if verbose:
            print("Begin fitting individual", indiv_i, "of", len(population))

        _input = copy.deepcopy(input)
        use_saved_estimator = kwargs.get("use_saved_estimator", False)

        # check for 'use_saved_estimator'
        if use_saved_estimator and getattr(indiv, "input", None) is not None:
            _input["estimator"] = indiv.input["estimator"]

        estimator = make_estimator(_input)

        indiv.fitness.values = cross_val_score(estimator,
                                               fdata.weighted_X,
                                               indiv,
                                               y=fdata.weighted_y,
                                               scoring=fdata.scoring,
                                               cv=fdata.cv,
                                               penalty=fdata.penalty)
        add_individual_detail(indiv, estimator, fdata, _input, selector=None)

        if verbose:
            print("  DONE\n")

    if verbose:
        print("Result:")
        print_halloffame(population)

    if hall is not None:
        hall.update(population)

    return population
